# Author: Carsten Sachse 08-Jun-2011
# Copyright: EMBL (2010 - 2018)
# License: see license.txt for details
from EMAN2 import EMData, Util, Transform, Vec2f, Reconstructors
from spring.micprgs.scansplit import Micrograph
from spring.segment2d.segment import Segment
from spring.segment3d.segclassreconstruct_prep import SegClassReconstructCylinderMask
from tabulate import tabulate
from utilities import compose_transform2, model_circle, get_sym
import numpy as np


class SegClassReconstructAssist(SegClassReconstructCylinderMask):                
    def compute_transform_with_aligned_helix_segment(self, phi, theta, psi):
        transform = Transform({'type':'spider', 'phi':phi, 'theta':theta, 'psi':psi})
        transform_helical_origin = Transform({'type':'spider', 'phi':0.0, 'theta':90.0, 'psi':270.0})
        transposed_helical_origin = transform_helical_origin.transpose()
        composite_transform = transform * transposed_helical_origin
        RABTeuler = composite_transform.get_rotation('spider')
        
        return RABTeuler


    def compute_helical_inplane_rotation_from_Euler_angles(self, phi, theta, psi):
        """
        >>> from spring.segment3d.segclassreconstruct import SegClassReconstruct
        >>> SegClassReconstruct().compute_helical_inplane_rotation_from_Euler_angles(180, 90, 270)
        0.0
        >>> SegClassReconstruct().compute_helical_inplane_rotation_from_Euler_angles(0, 90, 270)
        0.0
        >>> SegClassReconstruct().compute_helical_inplane_rotation_from_Euler_angles(33, 90, 334)
        64.00000006504217
        >>> SegClassReconstruct().compute_helical_inplane_rotation_from_Euler_angles(180, 90, 240)
        329.999999554708
        >>> SegClassReconstruct().compute_helical_inplane_rotation_from_Euler_angles(150.0, 94.0, 315)
        45.0
        >>> SegClassReconstruct().compute_helical_inplane_rotation_from_Euler_angles(150.0, 86.0, 315)
        225.0
        >>> SegClassReconstruct().compute_helical_inplane_rotation_from_Euler_angles(170.0, 90.0, 315)
        45.00000000000006
        """
        RABTeuler = self.compute_transform_with_aligned_helix_segment(phi, 90.0, psi)
        RABTphi = RABTeuler['phi']
        RABTpsi = RABTeuler['psi']
        
#        delta_inplane_rotation = (RABTpsi + RABTphi) % 360
        if theta >= 90:
            delta_inplane_rotation = (RABTpsi + RABTphi) % 360
        else:
            delta_inplane_rotation = (RABTpsi + RABTphi + 180) % 360
        
        return delta_inplane_rotation
    

    def compute_point_where_normal_and_parallel_line_intersect(self, shift_x, shift_y, inplane_angle):
        """
        >>> from spring.segment3d.segclassreconstruct import SegClassReconstruct
        >>> s = SegClassReconstruct()
        >>> s.compute_point_where_normal_and_parallel_line_intersect(3, 1, 0)
        (3.0, -1.8369701987210297e-16)
        >>> s.compute_point_where_normal_and_parallel_line_intersect(3, 1, 45)
        (2.0000000000000004, 2.0)
        >>> s.compute_point_where_normal_and_parallel_line_intersect(3, 1, -45)
        (0.99999999999999978, -0.99999999999999989)
        >>> s.compute_point_where_normal_and_parallel_line_intersect(3, 1, 90)
        (1.7449489701899453e-12, 1.0000000000052349)
        """
        if inplane_angle == 90.0 or inplane_angle == 270.0:
            inplane_angle -= 1e-10
            
        if 90.0 + inplane_angle == 0:
            slope = np.tan(np.deg2rad(360.0))
        else:
            slope = np.tan(np.deg2rad(90.0 + inplane_angle))
            
        t = shift_y - slope * shift_x 
        
        point_x = -t / (1/slope + slope)
        point_y = -point_x / slope 
        
        return point_x, point_y
    
        
    def compute_distances_to_helical_axis(self, shift_x, shift_y, inplane_angle):
        """
        >>> from spring.segment3d.segclassreconstruct import SegClassReconstruct
        >>> SegClassReconstruct().compute_distances_to_helical_axis(22, 2, 0)
        (-22.0, 2.0000000000000013)
        >>> SegClassReconstruct().compute_distances_to_helical_axis(22, 2, 180)
        (22.0, -2.000000000000004)
        >>> SegClassReconstruct().compute_distances_to_helical_axis(22, 2, 90)
        (-2.0000000000383888, -21.999999999996511)
        >>> SegClassReconstruct().compute_distances_to_helical_axis(3, 1, -45)
        (-1.4142135623730949, 2.8284271247461898)
        >>> SegClassReconstruct().compute_distances_to_helical_axis(3, 1, 45)
        (-2.8284271247461903, -1.4142135623730947)
        >>> SegClassReconstruct().compute_distances_to_helical_axis(22, 2, 20)
        (-21.357277943941323, -5.6450579115928949)
        >>> s = SegClassReconstruct()
        >>> hx, hy = s.compute_distances_to_helical_axis(22, 2, 180)
        >>> SegClassReconstruct().compute_distances_to_helical_axis(hx, hy, 180)
        (22.0, 2.0)
        """
        
        point_x, point_y = self.compute_point_where_normal_and_parallel_line_intersect(shift_x, shift_y, inplane_angle)
        dist_x, dst_y = Segment().rotate_coordinates_by_angle(-point_x, -point_y, inplane_angle)
        
        centered_x = shift_x - point_x
        centered_y = shift_y - point_y
        dst_x, dist_y = Segment().rotate_coordinates_by_angle(centered_x, centered_y, inplane_angle)
        
        return dist_x, dist_y
    
    
    def prepare_symmetry_computation(self, alignment_parameters, each_symmetry_pair, pixelsize, symmetry_views_count):
        helical_rise = each_symmetry_pair[0] / pixelsize
        helical_rotation = each_symmetry_pair[1]
        symmetry_views = np.arange(-symmetry_views_count / 2, symmetry_views_count / 2, dtype='float64')
        
        x_norm, y_parallel = self.compute_distances_to_helical_axis(alignment_parameters.x_shift,
        alignment_parameters.y_shift, alignment_parameters.inplane_angle)
        
        if helical_rise != 0:
            multiple_off_center = np.round(y_parallel / helical_rise)
        else:
            multiple_off_center = np.float(0)
            
        symmetry_views += multiple_off_center
        out_of_plane_correction = np.sin(np.deg2rad(alignment_parameters.theta))
        
        return symmetry_views, helical_rotation, out_of_plane_correction, helical_rise

        
    def compute_straight_phi_and_yshift(self, each_symmetry_view, phi, helical_rotation, out_of_plane_correction,
    helical_rise):
        symmetrized_phi = (phi + each_symmetry_view * helical_rotation) % 360
        straight_yshift = out_of_plane_correction * each_symmetry_view * helical_rise
        
        return straight_yshift, symmetrized_phi


    def compute_sx_sy_from_shifts_normal_and_parallel_to_helix_axis_with_inplane_angle(self, helix_shift_x,
    helix_shift_y, inplane_angle):
        """
        >>> from spring.segment3d.segclassreconstruct import SegClassReconstruct
        >>> hx, hy = SegClassReconstruct().compute_distances_to_helical_axis(3, 1, -45)
        >>> s = SegClassReconstruct()
        >>> s.compute_sx_sy_from_shifts_normal_and_parallel_to_helix_axis_with_inplane_angle(hx, hy, -45)
        (2.9999999999999996, 1.0000000000000002)
        >>> s.compute_sx_sy_from_shifts_normal_and_parallel_to_helix_axis_with_inplane_angle(3, 0, -45)
        (-2.1213203435596428, 2.1213203435596424)
        >>> hx, hy = SegClassReconstruct().compute_distances_to_helical_axis(22, 2, 20)
        >>> s.compute_sx_sy_from_shifts_normal_and_parallel_to_helix_axis_with_inplane_angle(hx, hy, 20)
        (22.0, 2.0)
        """
        point_x, point_y = Segment().rotate_coordinates_by_angle(helix_shift_x, 0, -inplane_angle)
        helix_y_from_point = helix_shift_y - point_y
        sx, sy = Segment().rotate_coordinates_by_angle(-point_x, helix_y_from_point, -inplane_angle, -point_x, -point_y)
        
        return sx, sy
        
    
    def correct_straight_symmetry_parameters_for_in_plane_rotation_and_shifts_old(self, inplane_angle, x_shift, y_shift,
    straight_yshift):
#        x_contribution, y_contribution = self.compute_sx_sy_from_shifts_normal_and_parallel_to_helix_axis_with_inplane_angle(0.0, straight_yshift, inplane_angle)

        new_rot, x_contribution, y_contribution, scale = compose_transform2(0.0, 0, straight_yshift, 1.0, inplane_angle,
        0.0, 0.0, 1.0)
        
        total_symmetrized_xshift = x_contribution + x_shift
        total_symmetrized_yshift = y_contribution + y_shift
        
        return total_symmetrized_xshift, total_symmetrized_yshift
    
    
    def correct_straight_symmetry_parameters_for_in_plane_rotation_and_shifts(self, inplane_angle, x_shift, y_shift,
    straight_yshift):
        
        x_norm, y_parallel = self.compute_distances_to_helical_axis(x_shift, y_shift, inplane_angle)
        
        total_symmetrized_xshift, total_symmetrized_yshift = \
        self.compute_sx_sy_from_shifts_normal_and_parallel_to_helix_axis_with_inplane_angle(x_norm, y_parallel + \
        straight_yshift, inplane_angle)
        
        return -total_symmetrized_xshift, -total_symmetrized_yshift
    

    def compute_helical_symmetry_related_views(self, alignment_parameters, each_symmetry_pair, pixelsize,
    symmetry_views_count):
        """
        #. positive y-shift for straight helix
        #. negative y-shift for straight helix
        #. negative y-shift including x-shift with flipped helix
        #. helix perpendicular to 'straight'
        #. id3 but respective x and y shifts swapped
        #. test angle of 45 degrees
        #. test angle of 60 degrees
        #. Out-of-plane tilt will affect helical rise in image plane:
        #. If straight y-shift is smaller than helical rise, no action need to taken
        #. If straight y-shift is larger than half helical rise will be reduced:
        
        >>> from spring.segment3d.segclassreconstruct import SegClassReconstruct
        >>> s = SegClassReconstruct()
        >>> rec_parameters = s.make_named_tuple_for_reconstruction()
        >>> r_info = rec_parameters(0, 0, 0, 90, 270, 0, 0, 0.0, 0, 22)
        >>> s.compute_helical_symmetry_related_views(r_info, (10, 50), 10, 10)
        (array([[   0.,  110.,   90.,  270.,   -0.,    5.,    0.,   22.],
               [   0.,  160.,   90.,  270.,   -0.,    4.,    0.,   22.],
               [   0.,  210.,   90.,  270.,   -0.,    3.,    0.,   22.],
               [   0.,  260.,   90.,  270.,   -0.,    2.,    0.,   22.],
               [   0.,  310.,   90.,  270.,   -0.,    1.,    0.,   22.],
               [   0.,    0.,   90.,  270.,   -0.,   -0.,    0.,   22.],
               [   0.,   50.,   90.,  270.,   -0.,   -1.,    0.,   22.],
               [   0.,  100.,   90.,  270.,   -0.,   -2.,    0.,   22.],
               [   0.,  150.,   90.,  270.,   -0.,   -3.,    0.,   22.],
               [   0.,  200.,   90.,  270.,   -0.,   -4.,    0.,   22.]]), 0.0)

        >>> r_info = rec_parameters(0, 0, 0, 90, 270, 0.5, 3.0, 0.0, 0, 22)
        >>> s.compute_helical_symmetry_related_views(r_info, (10, 50), 10, 10)
        (array([[   0. ,  260. ,   90. ,  270. ,    0.5,    5. ,    0. ,   22. ],
               [   0. ,  310. ,   90. ,  270. ,    0.5,    4. ,    0. ,   22. ],
               [   0. ,    0. ,   90. ,  270. ,    0.5,    3. ,    0. ,   22. ],
               [   0. ,   50. ,   90. ,  270. ,    0.5,    2. ,    0. ,   22. ],
               [   0. ,  100. ,   90. ,  270. ,    0.5,    1. ,    0. ,   22. ],
               [   0. ,  150. ,   90. ,  270. ,    0.5,   -0. ,    0. ,   22. ],
               [   0. ,  200. ,   90. ,  270. ,    0.5,   -1. ,    0. ,   22. ],
               [   0. ,  250. ,   90. ,  270. ,    0.5,   -2. ,    0. ,   22. ],
               [   0. ,  300. ,   90. ,  270. ,    0.5,   -3. ,    0. ,   22. ],
               [   0. ,  350. ,   90. ,  270. ,    0.5,   -4. ,    0. ,   22. ]]), 0.0)

        >>> r_info = rec_parameters(0, 0, 0, 90, 270, 0, 0, 0.0, 0, 22)
        >>> s.compute_helical_symmetry_related_views(r_info, (10, 50), 10, 2)
        (array([[   0.,  310.,   90.,  270.,   -0.,    1.,    0.,   22.],
               [   0.,    0.,   90.,  270.,   -0.,   -0.,    0.,   22.]]), 0.0)
              
        >>> r_info = rec_parameters(3, 3, 0, 90, 90, 2.5, -3.5, 180.0, 0, 22)
        >>> s.compute_helical_symmetry_related_views(r_info, (10, 50), 10, 10)
        (array([[   3. ,  260. ,   90. ,   90. ,    2.5,   -5.5,    0. ,   22. ],
               [   3. ,  310. ,   90. ,   90. ,    2.5,   -4.5,    0. ,   22. ],
               [   3. ,    0. ,   90. ,   90. ,    2.5,   -3.5,    0. ,   22. ],
               [   3. ,   50. ,   90. ,   90. ,    2.5,   -2.5,    0. ,   22. ],
               [   3. ,  100. ,   90. ,   90. ,    2.5,   -1.5,    0. ,   22. ],
               [   3. ,  150. ,   90. ,   90. ,    2.5,   -0.5,    0. ,   22. ],
               [   3. ,  200. ,   90. ,   90. ,    2.5,    0.5,    0. ,   22. ],
               [   3. ,  250. ,   90. ,   90. ,    2.5,    1.5,    0. ,   22. ],
               [   3. ,  300. ,   90. ,   90. ,    2.5,    2.5,    0. ,   22. ],
               [   3. ,  350. ,   90. ,   90. ,    2.5,    3.5,    0. ,   22. ]]), 180.0)

        >>> r_info = rec_parameters(2, 2, 0, 90, 180, 2.5, -3.5, 270.0, 0, 22)
        >>> s.compute_helical_symmetry_related_views(r_info, (10, 50), 10, 10)
        (array([[   2. ,  260. ,   90. ,  180. ,    4.5,   -3.5,    0. ,   22. ],
               [   2. ,  310. ,   90. ,  180. ,    3.5,   -3.5,    0. ,   22. ],
               [   2. ,    0. ,   90. ,  180. ,    2.5,   -3.5,    0. ,   22. ],
               [   2. ,   50. ,   90. ,  180. ,    1.5,   -3.5,    0. ,   22. ],
               [   2. ,  100. ,   90. ,  180. ,    0.5,   -3.5,    0. ,   22. ],
               [   2. ,  150. ,   90. ,  180. ,   -0.5,   -3.5,    0. ,   22. ],
               [   2. ,  200. ,   90. ,  180. ,   -1.5,   -3.5,    0. ,   22. ],
               [   2. ,  250. ,   90. ,  180. ,   -2.5,   -3.5,    0. ,   22. ],
               [   2. ,  300. ,   90. ,  180. ,   -3.5,   -3.5,    0. ,   22. ],
               [   2. ,  350. ,   90. ,  180. ,   -4.5,   -3.5,    0. ,   22. ]]), 270.0)

        >>> r_info = rec_parameters(4, 4, 0, 90, 180, -2.5, 3.5, 270.0, 0, 22)
        >>> s.compute_helical_symmetry_related_views(r_info, (10, 50), 10, 10)
        (array([[   4. ,  320. ,   90. ,  180. ,    5.5,    3.5,    0. ,   22. ],
               [   4. ,   10. ,   90. ,  180. ,    4.5,    3.5,    0. ,   22. ],
               [   4. ,   60. ,   90. ,  180. ,    3.5,    3.5,    0. ,   22. ],
               [   4. ,  110. ,   90. ,  180. ,    2.5,    3.5,    0. ,   22. ],
               [   4. ,  160. ,   90. ,  180. ,    1.5,    3.5,    0. ,   22. ],
               [   4. ,  210. ,   90. ,  180. ,    0.5,    3.5,    0. ,   22. ],
               [   4. ,  260. ,   90. ,  180. ,   -0.5,    3.5,    0. ,   22. ],
               [   4. ,  310. ,   90. ,  180. ,   -1.5,    3.5,    0. ,   22. ],
               [   4. ,    0. ,   90. ,  180. ,   -2.5,    3.5,    0. ,   22. ],
               [   4. ,   50. ,   90. ,  180. ,   -3.5,    3.5,    0. ,   22. ]]), 270.0)

        >>> r_info = rec_parameters(5, 5, 0, 90, 45, 1.0, 1.0, 135.0, 0, 22)
        >>> s.compute_helical_symmetry_related_views(r_info, (10, 50), 10, 10)
        (array([[  5.00000000e+00,   6.00000000e+01,   9.00000000e+01,
                  4.50000000e+01,  -3.24264069e+00,  -3.24264069e+00,
                  0.00000000e+00,   2.20000000e+01],
               [  5.00000000e+00,   1.10000000e+02,   9.00000000e+01,
                  4.50000000e+01,  -2.53553391e+00,  -2.53553391e+00,
                  0.00000000e+00,   2.20000000e+01],
               [  5.00000000e+00,   1.60000000e+02,   9.00000000e+01,
                  4.50000000e+01,  -1.82842712e+00,  -1.82842712e+00,
                  0.00000000e+00,   2.20000000e+01],
               [  5.00000000e+00,   2.10000000e+02,   9.00000000e+01,
                  4.50000000e+01,  -1.12132034e+00,  -1.12132034e+00,
                  0.00000000e+00,   2.20000000e+01],
               [  5.00000000e+00,   2.60000000e+02,   9.00000000e+01,
                  4.50000000e+01,  -4.14213562e-01,  -4.14213562e-01,
                  0.00000000e+00,   2.20000000e+01],
               [  5.00000000e+00,   3.10000000e+02,   9.00000000e+01,
                  4.50000000e+01,   2.92893219e-01,   2.92893219e-01,
                  0.00000000e+00,   2.20000000e+01],
               [  5.00000000e+00,   0.00000000e+00,   9.00000000e+01,
                  4.50000000e+01,   1.00000000e+00,   1.00000000e+00,
                  0.00000000e+00,   2.20000000e+01],
               [  5.00000000e+00,   5.00000000e+01,   9.00000000e+01,
                  4.50000000e+01,   1.70710678e+00,   1.70710678e+00,
                  0.00000000e+00,   2.20000000e+01],
               [  5.00000000e+00,   1.00000000e+02,   9.00000000e+01,
                  4.50000000e+01,   2.41421356e+00,   2.41421356e+00,
                  0.00000000e+00,   2.20000000e+01],
               [  5.00000000e+00,   1.50000000e+02,   9.00000000e+01,
                  4.50000000e+01,   3.12132034e+00,   3.12132034e+00,
                  0.00000000e+00,   2.20000000e+01]]), 135.0)
        
        >>> r_info = rec_parameters(6, 3, 0, 90, 60, 1, 1, 150.0, 0, 22)
        >>> s.compute_helical_symmetry_related_views(r_info, (10, 50), 10, 10)
        (array([[  6.00000000e+00,   6.00000000e+01,   9.00000000e+01,
                  6.00000000e+01,  -2.00000000e+00,  -4.19615242e+00,
                  0.00000000e+00,   2.20000000e+01],
               [  6.00000000e+00,   1.10000000e+02,   9.00000000e+01,
                  6.00000000e+01,  -1.50000000e+00,  -3.33012702e+00,
                  0.00000000e+00,   2.20000000e+01],
               [  6.00000000e+00,   1.60000000e+02,   9.00000000e+01,
                  6.00000000e+01,  -1.00000000e+00,  -2.46410162e+00,
                  0.00000000e+00,   2.20000000e+01],
               [  6.00000000e+00,   2.10000000e+02,   9.00000000e+01,
                  6.00000000e+01,  -5.00000000e-01,  -1.59807621e+00,
                  0.00000000e+00,   2.20000000e+01],
               [  6.00000000e+00,   2.60000000e+02,   9.00000000e+01,
                  6.00000000e+01,  -3.33066907e-16,  -7.32050808e-01,
                  0.00000000e+00,   2.20000000e+01],
               [  6.00000000e+00,   3.10000000e+02,   9.00000000e+01,
                  6.00000000e+01,   5.00000000e-01,   1.33974596e-01,
                  0.00000000e+00,   2.20000000e+01],
               [  6.00000000e+00,   0.00000000e+00,   9.00000000e+01,
                  6.00000000e+01,   1.00000000e+00,   1.00000000e+00,
                  0.00000000e+00,   2.20000000e+01],
               [  6.00000000e+00,   5.00000000e+01,   9.00000000e+01,
                  6.00000000e+01,   1.50000000e+00,   1.86602540e+00,
                  0.00000000e+00,   2.20000000e+01],
               [  6.00000000e+00,   1.00000000e+02,   9.00000000e+01,
                  6.00000000e+01,   2.00000000e+00,   2.73205081e+00,
                  0.00000000e+00,   2.20000000e+01],
               [  6.00000000e+00,   1.50000000e+02,   9.00000000e+01,
                  6.00000000e+01,   2.50000000e+00,   3.59807621e+00,
                  0.00000000e+00,   2.20000000e+01]]), 150.0)
                  
        >>> s.compute_helical_symmetry_related_views(r_info, (10, 50), 10, 10)
        (array([[  6.00000000e+00,   6.00000000e+01,   9.00000000e+01,
                  6.00000000e+01,  -2.00000000e+00,  -4.19615242e+00,
                  0.00000000e+00,   2.20000000e+01],
               [  6.00000000e+00,   1.10000000e+02,   9.00000000e+01,
                  6.00000000e+01,  -1.50000000e+00,  -3.33012702e+00,
                  0.00000000e+00,   2.20000000e+01],
               [  6.00000000e+00,   1.60000000e+02,   9.00000000e+01,
                  6.00000000e+01,  -1.00000000e+00,  -2.46410162e+00,
                  0.00000000e+00,   2.20000000e+01],
               [  6.00000000e+00,   2.10000000e+02,   9.00000000e+01,
                  6.00000000e+01,  -5.00000000e-01,  -1.59807621e+00,
                  0.00000000e+00,   2.20000000e+01],
               [  6.00000000e+00,   2.60000000e+02,   9.00000000e+01,
                  6.00000000e+01,  -3.33066907e-16,  -7.32050808e-01,
                  0.00000000e+00,   2.20000000e+01],
               [  6.00000000e+00,   3.10000000e+02,   9.00000000e+01,
                  6.00000000e+01,   5.00000000e-01,   1.33974596e-01,
                  0.00000000e+00,   2.20000000e+01],
               [  6.00000000e+00,   0.00000000e+00,   9.00000000e+01,
                  6.00000000e+01,   1.00000000e+00,   1.00000000e+00,
                  0.00000000e+00,   2.20000000e+01],
               [  6.00000000e+00,   5.00000000e+01,   9.00000000e+01,
                  6.00000000e+01,   1.50000000e+00,   1.86602540e+00,
                  0.00000000e+00,   2.20000000e+01],
               [  6.00000000e+00,   1.00000000e+02,   9.00000000e+01,
                  6.00000000e+01,   2.00000000e+00,   2.73205081e+00,
                  0.00000000e+00,   2.20000000e+01],
               [  6.00000000e+00,   1.50000000e+02,   9.00000000e+01,
                  6.00000000e+01,   2.50000000e+00,   3.59807621e+00,
                  0.00000000e+00,   2.20000000e+01]]), 150.0)
                  
        >>> s.compute_helical_symmetry_related_views(r_info, (10, 50), 10, 10)
        (array([[  6.00000000e+00,   6.00000000e+01,   9.00000000e+01,
                  6.00000000e+01,  -2.00000000e+00,  -4.19615242e+00,
                  0.00000000e+00,   2.20000000e+01],
               [  6.00000000e+00,   1.10000000e+02,   9.00000000e+01,
                  6.00000000e+01,  -1.50000000e+00,  -3.33012702e+00,
                  0.00000000e+00,   2.20000000e+01],
               [  6.00000000e+00,   1.60000000e+02,   9.00000000e+01,
                  6.00000000e+01,  -1.00000000e+00,  -2.46410162e+00,
                  0.00000000e+00,   2.20000000e+01],
               [  6.00000000e+00,   2.10000000e+02,   9.00000000e+01,
                  6.00000000e+01,  -5.00000000e-01,  -1.59807621e+00,
                  0.00000000e+00,   2.20000000e+01],
               [  6.00000000e+00,   2.60000000e+02,   9.00000000e+01,
                  6.00000000e+01,  -3.33066907e-16,  -7.32050808e-01,
                  0.00000000e+00,   2.20000000e+01],
               [  6.00000000e+00,   3.10000000e+02,   9.00000000e+01,
                  6.00000000e+01,   5.00000000e-01,   1.33974596e-01,
                  0.00000000e+00,   2.20000000e+01],
               [  6.00000000e+00,   0.00000000e+00,   9.00000000e+01,
                  6.00000000e+01,   1.00000000e+00,   1.00000000e+00,
                  0.00000000e+00,   2.20000000e+01],
               [  6.00000000e+00,   5.00000000e+01,   9.00000000e+01,
                  6.00000000e+01,   1.50000000e+00,   1.86602540e+00,
                  0.00000000e+00,   2.20000000e+01],
               [  6.00000000e+00,   1.00000000e+02,   9.00000000e+01,
                  6.00000000e+01,   2.00000000e+00,   2.73205081e+00,
                  0.00000000e+00,   2.20000000e+01],
               [  6.00000000e+00,   1.50000000e+02,   9.00000000e+01,
                  6.00000000e+01,   2.50000000e+00,   3.59807621e+00,
                  0.00000000e+00,   2.20000000e+01]]), 150.0)
        
        """
        symmetry_alignment_parameters = []
        symmetry_views, helical_rotation, out_of_plane_correction, helical_rise = \
            self.prepare_symmetry_computation(alignment_parameters, each_symmetry_pair, pixelsize, symmetry_views_count)
        
        for each_symmetry_related_view in symmetry_views:
            straight_yshift, symmetrized_phi = self.compute_straight_phi_and_yshift(each_symmetry_related_view,
            alignment_parameters.phi, helical_rotation, out_of_plane_correction, helical_rise)
            
            total_symmetrized_xshift, total_symmetrized_yshift = \
            self.correct_straight_symmetry_parameters_for_in_plane_rotation_and_shifts(
            alignment_parameters.inplane_angle, -alignment_parameters.x_shift, -alignment_parameters.y_shift, 
            straight_yshift)
            
            symmetry_alignment_parameters.append(np.array([alignment_parameters.stack_id, symmetrized_phi,
            alignment_parameters.theta, alignment_parameters.psi, total_symmetrized_xshift, total_symmetrized_yshift,
            alignment_parameters.mirror, alignment_parameters.seg_ref_id], dtype=float))
            
        if symmetry_alignment_parameters != []:
            symmetry_alignment_parameters = np.vstack(symmetry_alignment_parameters)
        
        return symmetry_alignment_parameters, alignment_parameters.inplane_angle
    
    
    def setup_reconstructor(self, imgsize):
        fftvol = EMData()
        weight = EMData()
        #ds.humanize_bytes(s.compute_byte_size_of_image_stack(4 * 200,4 * 200,4 * 200)) = 1.91 GB
        if 0 < imgsize <= 200:
            npad = 4
        # ds.humanize_bytes(s.compute_byte_size_of_image_stack(3 * 267,3 * 267,3 * 267)) = 1.91GB
        elif 200 < imgsize <= 267:
            npad = 3
        #ds.humanize_bytes(s.compute_byte_size_of_image_stack(2 * 400,2 * 400,2 * 400)) = 1.91 GB
        elif 267 < imgsize <= 400:
            npad = 2
        elif imgsize > 400: 
            npad = 1.5
        params = {'size':imgsize, 'npad':npad, 'symmetry':'c1', 'fftvol':fftvol, 'weight':weight}
        r = Reconstructors.get('nn4', params)
        r.setup()
        
        return r, fftvol, weight
    

    def setup_transform_with_reduced_alignment_parameters(self, each_phi, each_theta, each_psi, each_x, each_y):
        transform_projection = Transform({'type':'spider', 'phi':each_phi, 'theta':each_theta, 'psi':each_psi})
        transform_projection.set_trans(Vec2f(-each_x, -each_y))
        
        return transform_projection
    

    def get_symmetry_transformations(self, point_symmetry):
        angs = get_sym(point_symmetry)
        
        sym_transformations = [Transform({'type':'spider', 'phi':each_phi, 'theta':each_theta, 'psi':each_psi}) for \
        each_phi, each_theta, each_psi in angs]
        
        return sym_transformations


    def get_Euler_angles_after_transformation(self, each_phi, each_theta, each_psi, each_x, each_y, each_sym_transform):
        current_transform = Transform({'type':'spider', 'phi':each_phi, 'theta':each_theta, 'psi':each_psi, 
                                       'tx': each_x, 'ty': each_y})
        transform_projection = current_transform * each_sym_transform
        trans = transform_projection.get_params('spider')
        
        return transform_projection, trans['phi'], trans['theta'], trans['psi'], trans['tx'], trans['ty']
    
    
    def compute_point_group_symmetry_related_views(self, sym_transformations, align_params, inplane=False):
        symmetry_alignment_parameters = []
        for (each_stack_id, each_phi, each_theta, each_psi, each_x, each_y, each_mirror, each_ref_id) in align_params:
            for each_sym_transform in sym_transformations:
                if inplane:
                    each_psi = 270.0
                transform_projection, trans_phi, trans_theta, trans_psi, trans_x, trans_y = \
                self.get_Euler_angles_after_transformation(each_phi, each_theta, each_psi, each_x, each_y, 
                each_sym_transform)
                
                symmetry_alignment_parameters.append(np.array([each_stack_id, trans_phi, trans_theta, trans_psi, trans_x,
                trans_y, each_mirror, each_ref_id], dtype=float))
        
        if symmetry_alignment_parameters != []:
            symmetry_alignment_parameters = np.vstack(symmetry_alignment_parameters)
        
        return symmetry_alignment_parameters
        

    def convert_alignment_parameters_to_symmetry_alignment_parameters(self, each_param):
        sym_align_parameters = np.vstack([[each_param.stack_id, each_param.phi, each_param.theta, each_param.psi,
        each_param.x_shift, each_param.y_shift, each_param.mirror, each_param.seg_ref_id]])
        
        return sym_align_parameters, each_param.inplane_angle
    

    def compute_all_symmetry_related_views(self, each_symmetry_pair, pixelsize, symmetry_views_count, point_symmetry,
    each_alignment_parameters):
        """
        >>> from spring.segment3d.segclassreconstruct import SegClassReconstruct
        >>> s = SegClassReconstruct()
        >>> rec_parameters = s.make_named_tuple_for_reconstruction()
        >>> r_info = rec_parameters(0, 0, 0, 90, 270, 0, 0, 0.0, 0, 22)
        >>> a, b, c = s.compute_all_symmetry_related_views((1.408, 22.03), 1.2, 2, 'c2', r_info)
        >>> a
        array([[   0.        ,  337.97000087,   90.        ,  270.        ,
                   0.        ,    1.17333329,    0.        ,   22.        ],
               [   0.        ,  157.97000087,   90.        ,  270.        ,
                   0.        ,    1.17333329,    0.        ,   22.        ],
               [   0.        ,    0.        ,   90.        ,  270.        ,
                   0.        ,    0.        ,    0.        ,   22.        ],
               [   0.        ,  180.        ,   90.        ,  270.        ,
                   0.        ,    0.        ,    0.        ,   22.        ]])
        """
        sym_transformations = self.get_symmetry_transformations(point_symmetry)
        
        helical_rise, helical_rotation = each_symmetry_pair
        if helical_rise == 0 and helical_rotation == 0:
            sym_align_params, inplane_angle = \
            self.convert_alignment_parameters_to_symmetry_alignment_parameters(each_alignment_parameters)
        else:
            sym_align_params, inplane_angle = self.compute_helical_symmetry_related_views(each_alignment_parameters,
            each_symmetry_pair, pixelsize, symmetry_views_count)
            
        log_sym_align_params = self.compute_point_group_symmetry_related_views(sym_transformations, sym_align_params)

        sym_align_params = self.compute_point_group_symmetry_related_views(sym_transformations, sym_align_params,
        inplane=True)
    
        return sym_align_params, inplane_angle, log_sym_align_params
    

    def compute_and_write_symmetry_related_views(self, rec_volume, image_stack, alignment_parameters,
    each_symmetry_pair, pixelsize, symmetry_views_count, trimmed_image_size, rec_stack, point_symmetry='c1', 
    mode='BP 3F'):
        
        if mode == 'BP RP':
            mask2d = model_circle(trimmed_image_size/2 - 2, trimmed_image_size, trimmed_image_size)  # SIRT works for squares only!
            mask2d = 1.0 - mask2d  # invert the mask to get average in corners
        total_symmetry_alignment_parameters = []
        Euler_angles_rec = []
        
        symmetry_image_single = EMData()
        symmetry_rec_loginfo = []
        rec_img_id = -1
        for each_alignment_parameters in alignment_parameters:
            each_stack_id = int(each_alignment_parameters[0])
            each_local_id = each_alignment_parameters[1]
            symmetry_image_single.read_image(image_stack, each_local_id)
            
            sym_align_params, inplane_angle, log_sym_align_params = \
            self.compute_all_symmetry_related_views(each_symmetry_pair, pixelsize, symmetry_views_count, point_symmetry,
            each_alignment_parameters)
            
            for each_sym_id, (each_stack_id, each_phi, each_theta, each_psi, each_x, each_y, each_mirror, each_ref_id) \
            in enumerate(sym_align_params):
                symmetry_image = symmetry_image_single.copy()
                
                trans_symmetry_image = Segment().shift_and_rotate_image_gridding(symmetry_image, inplane_angle,
                each_x, each_y)
                
                transform_projection = Transform({'type':'spider', 'phi':each_phi, 'theta':each_theta, 'psi':each_psi,
                'tx':0.0, 'ty':0.0})
                
                if rec_stack is not None:
                    rec_image = Util.window(trans_symmetry_image, rec_stack.alignment_size, rec_stack.alignment_size, 
                                            1, 0, 0, 0)
                    rec_img_id += 1
                    rec_image.write_image(rec_stack.file_name, rec_img_id)
                    
#                trans_symmetry_image.append_image('test_rec.hdf')
                symmetry_image = Util.window(trans_symmetry_image, trimmed_image_size, trimmed_image_size, 1, 0, 0, 0)
                symmetry_image.set_attr('xform.projection', transform_projection)
                
                if mode == 'BP 3F':
                    rec_volume.insert_slice(symmetry_image, transform_projection)
                elif mode == 'BP RP':
                    stat = Micrograph().get_statistics_from_image(symmetry_image, mask2d)
                    symmetry_image = symmetry_image - stat.avg    #subtract the background average in the corners
                    
                    myparams = {'transform':transform_projection, 'anglelist':[each_phi, each_theta, each_psi],
                    'radius':(trimmed_image_size/2 - 2)}
                    
                    rec_volume += symmetry_image.backproject('chao', myparams)
                log_psi = log_sym_align_params[each_sym_id][3]
                symmetry_rec_loginfo += [[each_stack_id, each_phi, each_theta, log_psi, each_x, each_y]]
                
                total_symmetry_alignment_parameters.append(np.array([each_stack_id, each_phi, each_theta, log_psi,
                each_x, each_y, each_mirror, each_ref_id], dtype=float))
                
                Euler_angles_rec.append(np.array([rec_img_id, each_stack_id, each_phi, each_theta, each_psi, each_x,
                each_y, inplane_angle]))
                
        if total_symmetry_alignment_parameters != []:
            total_symmetry_alignment_parameters = np.vstack(total_symmetry_alignment_parameters)
            Euler_angles_rec = np.vstack(Euler_angles_rec)
        
        symmetry_rec_loginfo = tabulate(symmetry_rec_loginfo, ['stack_id', 'phi', 'theta', 'psi', 'x-shift', 'y-shift'])

        return rec_volume, total_symmetry_alignment_parameters, symmetry_rec_loginfo, Euler_angles_rec
    
    
    def project_from_volume_and_backproject(self, alignment_parameters, trimmed_image_size, xvol,
    pxvol):
        for each_image_parameters in alignment_parameters:
            for (each_stack_id, each_phi, each_theta, each_psi, each_x, each_y, each_ref_id) in each_image_parameters:
                
                transform_projection = \
                self.setup_transform_with_reduced_alignment_parameters(trimmed_image_size, each_phi,
                each_theta, each_psi, each_x, each_y)
                
                myparams = {'transform':transform_projection, 'anglelist':[each_phi, each_theta, each_psi],
                'radius':(trimmed_image_size/2 - 2)}
                
                data = xvol.project('chao', myparams) 
                pxvol += data.backproject('chao', myparams)
        
        return pxvol
    


# Author: Carsten Sachse 
# Copyright: EMBL (2010 - 2018)
# License: see license.txt for details
"""
Program to examine all excised in-plane rotated segments and compute their collapsed (1D) and 
2D power spectrum and width profile of helices
"""
from EMAN2 import Util, EMData, EMUtil, periodogram, EMNumPy
from filter import filt_gaussl
from fundamentals import image_decimate, fshift
from matplotlib import font_manager
from matplotlib.ticker import FuncFormatter
from morphology import threshold_maxval, threshold_to_zero, binarize
from spring.csinfrastr.csdatabase import SpringDataBase, base, SegmentTable, HelixTable
from spring.csinfrastr.csfeatures import Features
from spring.csinfrastr.cslogger import Logger
from spring.csinfrastr.csproductivity import DiagnosticPlot, Temporary, OpenMpi
from spring.csinfrastr.csreadinput import OptHandler
from spring.micprgs.scansplit import Micrograph
from spring.segment2d.segmentselect import SegmentSelect
from statistics import add_series, ccc
from tabulate import tabulate
from utilities import model_blank, model_circle
import matplotlib.cm as cm
import numpy as np
import os
import shutil

class SegmentExamPar(object):
    """
    Class to initiate default dictionary with input parameters including help and range values and 
    status dictionary
    """
    def __init__(self):
        # package/program identity
        self.package = 'emspring'
        self.progname = 'segmentexam'
        self.proginfo = __doc__
        self.code_files = [self.progname, self.progname + '_mpi']

        self.segmentexam_features = Features()
        self.feature_set = self.segmentexam_features.setup(self)
        
        self.define_parameters_and_their_properties()
        self.define_program_states()
    
    def define_parameters_and_their_properties(self):
        self.feature_set = self.segmentexam_features.set_inp_stack(self.feature_set)
        self.feature_set = self.segmentexam_features.set_output_plot(self.feature_set, self.progname + '_diag.pdf')
        
        self.feature_set = self.segmentexam_features.set_output_power_spectrum(self.feature_set)
        self.feature_set = self.segmentexam_features.set_enhance_power_option(self.feature_set)
        self.feature_set = self.segmentexam_features.set_output_enhanced_power_spectrum(self.feature_set)
        self.feature_set = self.segmentexam_features.set_pixelsize(self.feature_set)
        self.feature_set = self.segmentexam_features.set_helix_width(self.feature_set)
        self.feature_set = self.segmentexam_features.set_power_cutoff(self.feature_set)
        self.feature_set = self.set_compute_layer_line_correlation(self.feature_set)
        self.feature_set = self.set_layer_line_region(self.feature_set)
        self.feature_set = self.set_power_spectrum_reference(self.feature_set)
        self.feature_set = self.segmentexam_features.set_input_power_spectrum(self.feature_set,
        'Reference power spectrum', 'expert')
        
        self.feature_set = self.segmentexam_features.set_spring_path_segments(self.feature_set)

        self.feature_set = self.segmentexam_features.set_selection_criteria_from_segment_table(self.feature_set)
        
        self.feature_set = self.segmentexam_features.set_mpi(self.feature_set)
        self.feature_set = self.segmentexam_features.set_ncpus(self.feature_set)
        self.feature_set = self.segmentexam_features.set_temppath(self.feature_set)

    def define_program_states(self):
        self.feature_set.program_states['add_power_spectra_from_verticalized_stack']='Addition of in-plane rotated ' + \
        'power spectra'
        
        self.feature_set.program_states['determine_width']='Determine width of helix by projection along helical axis'
        
        self.feature_set.program_states['visualize_power_avg_and_width_analysis']='Visualization of width and power ' +\
        'spectra'


    def set_compute_layer_line_correlation(self, feature_set):
        inp9 = 'Compute layer-line correlation option'
        feature_set.parameters[inp9] = bool(False)
        feature_set.hints[inp9] = 'Compute layer line correlation of individual power spectra with average power ' + \
        'spectrum and store this correlation value in the Spring database (spring.db).'
        feature_set.level[inp9]='expert'
        
        return feature_set
    
    
    def set_power_spectrum_reference(self, feature_set):
        inp9 = 'Reference power spectrum'
        feature_set.parameters[inp9] = bool(False)
        feature_set.hints[inp9] = 'Choose whether to provide a reference power spectrum for layer-line ' + \
        'correlation. Make sure that the image dimensions agree with the input stack or with the binned input ' + \
        'stack. If not chosen, sum of power spectra from input stack will be used.'
        feature_set.level[inp9]='expert'
        feature_set.relatives[inp9]='Compute layer-line correlation option'
        
        return feature_set
    
    
    def set_layer_line_region(self, feature_set):
        inp7 = 'Layer-line region in 1/Angstrom'
        feature_set.parameters[inp7]=((0.03, 0.035))
        feature_set.hints[inp7]='Layer-line region to correlate individual power spectrum with sum of all power ' + \
        'spectra.'
        feature_set.properties[inp7]=feature_set.Range(0, 1, 0.001)
        feature_set.relatives[inp7]=(('Compute layer-line correlation option', 'Compute layer-line correlation option'))
        feature_set.level[inp7]='expert'
        
        return feature_set

        
class SegmentExamPower(object):
    """
    * Class that holds functions for examining segments from micrographs

    * __init__ Function to interpret multi-input parameters

    """
    def __init__(self, parset = None):
        self.log = Logger()
        if parset is not None:
            self.feature_set = parset
            p = self.feature_set.parameters

            self.infilestack = p['Image input stack']
            self.infile=self.infilestack
            self.outfile = p['Diagnostic plot']
            
            self.spring_path = p['spring.db file']
            
            self.power_img = p['Power spectrum output image']
            self.enhanced_power_option = p['Enhanced power spectrum option']
            self.power_enhanced_img = p['Enhanced power spectrum output image']
            self.pixelsize = p['Pixel size in Angstrom']
            self.helixwidth = p['Estimated helix width in Angstrom']
            self.helixwidthpix = int(round(self.helixwidth/self.pixelsize))
            self.rescutoff = p['Power spectrum resolution cutoff in 1/Angstrom']
            
            self.layer_ccc_option = p['Compute layer-line correlation option']
            self.res_ccc_range = p['Layer-line region in 1/Angstrom']
            self.power_reference = p['Reference power spectrum']
            self.power_input = p['Power spectrum input image']

            self = SegmentSelect().define_selection_parameters_from_segment_table(self, p)

            self.mpi_option = p['MPI option']
            self.cpu_count = p['Number of CPUs']
            self.temppath = p['Temporary directory']

            self.binfactor = max(1, int(round(1/(self.rescutoff*self.pixelsize*2))))

            self.stack = EMData()
            self.stack.read_image(self.infilestack, 0)
            self.segsizepix = self.stack.get_xsize()


    def bin_image_stack_by_binfactor(self, infilestack, binfactor, image_list=None, binned_stack=None):
        self.log.fcttolog()
        if image_list is None:
            image_list = list(range(EMUtil.get_image_count(infilestack)))
        if binned_stack is None:
            binned_stack = os.path.splitext(os.path.basename(infilestack))[0] + '{0}binned'.format(binfactor) + \
            os.path.splitext(infilestack)[-1]
        
        log_info = 'The following segments were binned by a factor {0} for further analysis and '.format(binfactor) + \
        'saved in {0}.\n(Local_id, Segment_id)\n'.format(binned_stack)
        segment = EMData()
        for each_local_id, each_seg_id in enumerate(image_list):
            segment.read_image(infilestack, each_seg_id)
            if binfactor > 1:
                segment = image_decimate(segment, binfactor, fit_to_fft=False)
            segment.write_image(binned_stack, each_local_id)
            log_info += '({0}, {1}) '.format(each_local_id, each_seg_id)
        
        self.log.ilog(log_info)
        segsizepix = segment.get_xsize()
        
        return binned_stack, segsizepix
    

    def apply_binfactor(self, binfactor, infilestack, segsizepix, helixwidthpix, pixelsize, image_list=None,
    outfile=None):
        """
        * Function to reduce stack and modify pixelsize according to desired binfactor

        #. Input: binfactor, infile stack to be binned, segment size (pixel), helix width (pixel) \
            pixelsize
        #. Output: binned stack, adjusted segment size (pixel), helix width (pixel), pixelsize 
        #. Usage: binned stack, segsizepix, helixwidth, pixelsize = apply_binfactor(binfactor, \
            infilestack, segsizepix, helixwidth, pixelsize)
            
        """

        binned_stack, segsizepix = self.bin_image_stack_by_binfactor(infilestack, binfactor, image_list, outfile)
            
        helixwidthpix = helixwidthpix/binfactor
        pixelsize = pixelsize * binfactor

        return binned_stack, segsizepix, helixwidthpix, pixelsize
    

    def enhance_power(self, avg_periodogram=None, pixelsize=None):
        """
        * Function to visually enhance power spectrum by compensating for decay of amplitude

        #. Input: power spectrum
        #. Output: enhanced power spectrum
        #. Usage: avg_periodogram_enhanced = enhance_power(avg_periodogram)
        
        """
        if avg_periodogram is None: 
            avg_periodogram = self.avg_periodogram
        if pixelsize is None: 
            pixelsize = self.pixelsize

        rotavg = avg_periodogram.rotavg_i()
        segment_size = avg_periodogram.get_xsize()
        mask = model_circle(segment_size/2, segment_size, segment_size)
        stat = Micrograph().get_statistics_from_image(rotavg, mask)
        if pixelsize <= 5:
            avg_periodogram_enhanced = avg_periodogram / (rotavg + 0.001*(stat.max - stat.avg))
        elif 5 < pixelsize < 10:
            avg_periodogram_enhanced = avg_periodogram / (rotavg + 0.01*(stat.max - stat.avg))
        elif pixelsize >= 10:
            avg_periodogram_enhanced = avg_periodogram / rotavg

        return avg_periodogram_enhanced
    

    def collapse_power(self, addpowimg):
        """
        * Function to project powerspectrum onto 1D plot to determine layer line position

        #. Input: power spectrum, segment size (pixel)
        #. Output: collapsed profile
        #. Usage: add1dimg = collapse_power(avg_periodogram)
        
        """
        segsizepix = addpowimg.get_xsize()

        circmask = model_circle(segsizepix - 1, segsizepix, segsizepix)
        maskaddpowimg = circmask * addpowimg
        halfaddpowimg = Util.window(maskaddpowimg, segsizepix, segsizepix/2, 1, 0,segsizepix/4, 0)
        img1dline = EMData()
        self.add1dimg = model_blank( segsizepix/2, 1, 1, 0)
        for eachCol in range(segsizepix/4, segsizepix*3/4, 1):
            img1dline = halfaddpowimg.get_col(eachCol)
            self.add1dimg += img1dline

        return self.add1dimg
    

    def project_helix(self, seg=None):
        """
        * Function to project image along helical axis by adding rows of image

        #. Input: segment
        #. Output: projected profile
        #. Usage: rowsaddimg = project_helix(seg)
        
        """
        if seg is None: 
            seg = self.seg

        number_of_rows = seg.get_ysize()
        self.rowsaddimg = model_blank(seg.get_xsize(), 1, 1, 0)
        for eachRow in range(number_of_rows):
            self.rowsaddimg += seg.get_row(eachRow)

        return self.rowsaddimg
    
    def project_normal_to_helix(self, seg=None):
        """
        * Function to project image perpendicular to helical axis by adding columns of image

        #. Input: segment
        #. Output: projected profile
        #. Usage: columnsaddimg = project_normal_to_helix(seg)
        
        """
        if seg is None: seg = self.seg

        number_of_columns = seg.get_xsize()
        self.columnsaddimg = model_blank(seg.get_ysize(), 1, 1, 0)
        for eachColumn in range(number_of_columns):
            self.columnsaddimg += seg.get_col(eachColumn)

        return self.columnsaddimg


class SegmentExamMask(SegmentExamPower):

    def limit_width_falloff_to_available_pixels_outside_binary_mask(self, helix_width_in_pixel, helix_height_in_pixel,
    segment_size_in_pixel, width_falloff):
        if (2 * segment_size_in_pixel * width_falloff + helix_width_in_pixel) > segment_size_in_pixel:
            width_falloff = float((segment_size_in_pixel - helix_width_in_pixel) / 2) / segment_size_in_pixel
        if helix_width_in_pixel >= segment_size_in_pixel:
            width_falloff = float((segment_size_in_pixel - helix_height_in_pixel) / 2) / segment_size_in_pixel
            
        return width_falloff
    

    def insure_mirror_symmetry_of_mask_parameters(self, helix_width_in_pixel, helix_height_in_pixel):
        helix_width_is_odd = (helix_width_in_pixel) % 2
        if helix_width_is_odd == 1:
            helix_width_in_pixel += 1
        helix_height_is_odd = (helix_height_in_pixel) % 2
        if helix_height_is_odd:
            helix_height_in_pixel += 1
            
        return helix_width_in_pixel, helix_height_in_pixel


    def make_binary_shape_mask(self, helix_width_in_pixel, helix_height_in_pixel, segment_size_in_pixel):
        if helix_width_in_pixel > segment_size_in_pixel and helix_height_in_pixel < segment_size_in_pixel:
            innermask = model_blank(int(segment_size_in_pixel), int(round(helix_height_in_pixel)), 1, 1)
            error_message = 'Helix is wider than specified segment size. Masking only in helix height.'
            self.log.wlog(error_message)
        elif helix_width_in_pixel > segment_size_in_pixel and helix_height_in_pixel > segment_size_in_pixel:
            innermask = model_blank(int(segment_size_in_pixel), int(round(segment_size_in_pixel)), 1, 1)
            error_message = 'Helix is wider and higher than specified segment size. No effective masking.'
            self.log.wlog(error_message)
        elif helix_width_in_pixel < segment_size_in_pixel and helix_height_in_pixel > segment_size_in_pixel:
            innermask = model_blank(int(helix_width_in_pixel), int(round(segment_size_in_pixel)), 1, 1)
            error_message = 'Helix is higher than specified segment size. Masking only in helix width.'
            self.log.wlog(error_message)
        else:
            innermask = model_blank(int(round(helix_width_in_pixel)), int(round(helix_height_in_pixel)), 1, 1)
        mask = Util.pad(innermask, segment_size_in_pixel, segment_size_in_pixel, 1, 0, 0, 0, '0')
        
        return mask


    def add_smooth_gaussian_falloff_to_edge_of_binary_mask(self, segment_size_in_pixel, width_falloff, binary_mask):
#        """
#        >>> from spring.segment2d.segmentexam import SegmentExam
#        >>> circ = model_circle(3, 20, 20)
#        >>> mask = SegmentExam().add_smooth_gaussian_falloff_to_edge_of_binary_mask(20, 0.1, circ)
#        >>> np.copy(EMNumPy.em2numpy(mask.get_row(10)))
#        """
        gauss_edge_width = max(1, int(round(segment_size_in_pixel * width_falloff)))
        kernel_width = gauss_edge_width * 2
        kernel_width_is_even = kernel_width % 2
        if kernel_width_is_even == 0:
            kernel_width += 1
        smooth_mask = filt_gaussl(binary_mask, 1/(2*np.pi*gauss_edge_width))
        smooth_mask += binary_mask
        try:
            smooth_mask = threshold_to_zero(smooth_mask, 0.075)
            smooth_mask = threshold_maxval(smooth_mask, 0.5)
        except RuntimeError:
            pass
        smooth_mask = smooth_mask * 2
        
        return smooth_mask
    

    def generate_falloff_line(self, segment_size_in_pixel, width_falloff):
        """
        >>> from spring.segment2d.segmentexam import SegmentExam
        >>> s = SegmentExam()
        >>> s.generate_falloff_line(20, 0.5)
        (array([ 1.        ,  0.97488286,  0.90205491,  0.78883309,  0.64659262,
                0.48962419,  0.33369821,  0.19448033,  0.08595758,  0.01903308]), 10)
        >>> s.generate_falloff_line(20, 0)
        (array([], dtype=float64), 0)
        """
        if width_falloff > 0:
            falloff_len = max(2, int(width_falloff * segment_size_in_pixel))
        else:
            falloff_len = 0

        falloff_line = 0.5 * np.cos(np.arange(falloff_len) / (falloff_len / 10.0 * np.pi)) + 0.5
        
        return falloff_line, falloff_len


    def generate_two_dee_cosine_falloff(self, falloff_line, max_dim):
        
        cos_field = np.zeros((len(falloff_line), max_dim))
        for every_row in np.arange(max_dim):
            cos_field[:,int(every_row)] = falloff_line
            
        return cos_field
    

    def generate_rectangular_mask_with_linear_falloffs(self, helix_width_in_pixel, helix_height_in_pixel,
    segment_size_in_pixel, width_falloff):
        falloff_line, falloff_len = self.generate_falloff_line(segment_size_in_pixel, width_falloff)
        
        cos_field = self.generate_two_dee_cosine_falloff(falloff_line, helix_height_in_pixel)
        rect = np.ones((helix_width_in_pixel, helix_height_in_pixel))
        mask = np.vstack([np.flipud(cos_field), rect, cos_field])
        mask = np.rot90(mask)
        
        cos_field = self.generate_two_dee_cosine_falloff(falloff_line, helix_width_in_pixel + 2 * falloff_len)
        mask_both = np.vstack([np.flipud(cos_field), mask, cos_field])
        
        return falloff_line, mask_both
    

    def generate_radial_falloff_gradient(self, falloff_line):
        falloff_len = len(falloff_line)
        img_dim = 2 * falloff_len
        img = model_blank(img_dim, img_dim)
        for each_radius, each_row in enumerate((falloff_line[:-1])):
            inner_circle = model_circle(each_radius, img_dim, img_dim)
            outer_circle = model_circle(each_radius + 1, img_dim, img_dim)
            ring = outer_circle - inner_circle
            img += (ring * each_row)
            
        square = np.copy(EMNumPy.em2numpy(img))
        
        upper_row, lower_row = np.hsplit(square, 2)
        third_quad, fourth_quadrant = np.vsplit(lower_row, 2)
        
        fourth_quadrant = fourth_quadrant[1:falloff_len, 1:falloff_len] 
        fq_range = fourth_quadrant.max() - fourth_quadrant.min()
        if fq_range != 0:
            fourth_quadrant = (fourth_quadrant - fourth_quadrant.min()) / (fourth_quadrant.max() - fourth_quadrant.min())
        else:
            fourth_quadrant = fourth_quadrant - fourth_quadrant.min()
        
        return fourth_quadrant
        
        
    def insert_radial_falloff_gradient_into_corners_of_rectangular_mask(self, radial_quadrant, falloff_line, rect_mask):
        falloff_len = len(falloff_line)
        rows, cols = np.shape(rect_mask)
        rect_mask[0:falloff_len - 1, 0:falloff_len - 1]=np.fliplr(np.flipud(radial_quadrant))
        rect_mask[rows - falloff_len + 1:rows, 0:falloff_len - 1]=np.fliplr(radial_quadrant)
        
        rect_mask[0:falloff_len - 1, cols - falloff_len + 1:cols]=np.flipud(radial_quadrant)
        rect_mask[rows - falloff_len + 1:rows, cols - falloff_len + 1:cols]=(radial_quadrant)
        
        central_row = rows / 2
        rect_mask[rows - falloff_len] = rect_mask[central_row]
        rect_mask[falloff_len - 1] = rect_mask[central_row]
        
        emmask = EMNumPy.numpy2em(np.copy(rect_mask))
        
        return emmask
    

    def pad_image_to_current_size(self, emmask, current_xsize, current_ysize):
        emmask = Util.pad(emmask, current_xsize, current_ysize, 1, 0, 0, 0, '0')
        return emmask


    def window_image_to_current_sizes(self, emmask, current_xsize, current_ysize):
        emmask = Util.window(emmask, current_xsize, current_ysize, 1, 0, 0, 0)
        return emmask

    def resize_mask_to_segment_dimensions(self, emmask, segment_size):
        current_xsize = emmask.get_xsize()
        current_ysize = emmask.get_ysize()
        
        if current_xsize < segment_size:
            current_xsize = segment_size
            emmask = self.pad_image_to_current_size(emmask, current_xsize, current_ysize)
        if current_ysize < segment_size:
            current_ysize = segment_size
            emmask = self.pad_image_to_current_size(emmask, current_xsize, current_ysize)
        if current_xsize > segment_size:
            current_xsize = segment_size
            emmask = self.window_image_to_current_sizes(emmask, current_xsize, current_ysize)
        if current_ysize > segment_size:
            current_ysize = segment_size
            emmask = self.window_image_to_current_sizes(emmask, current_xsize, current_ysize)
            
#        emmask.write_image('test.hdf')
        
        return emmask
    
    
    def compute_radial_average_from_line(self, falloff_line, falloff_len):
        center_x = falloff_len / 2
        center_y = falloff_len / 2
        
        xx, yy = np.mgrid[:falloff_len, :falloff_len]
        circle = np.sqrt((xx - center_x) ** 2 + (yy - center_y) ** 2)
        
        rad_avg_circle = np.zeros((falloff_len, falloff_len))
        for each_rad, each_val in enumerate(falloff_line):
            rad_avg_circle[((each_rad - 1) < circle) & (circle < (each_rad + 1))]=each_val 
            
        return rad_avg_circle
    

    def make_smooth_rectangular_mask(self, hel_width_pix, hel_height_pix, seg_size_pix, width_falloff=0.1):
        """
        >>> from spring.segment2d.segmentexam import SegmentExam
        >>> helixmask = SegmentExam().make_smooth_rectangular_mask(13, 30, 40)
        >>> mask_row = helixmask.get_row(20)
        >>> EMNumPy.em2numpy(mask_row)
        array([ 0.        ,  0.        ,  0.        ,  0.        ,  0.        ,
                0.        ,  0.        ,  0.        ,  0.        ,  0.        ,
                0.13561368,  0.4896242 ,  0.84986573,  1.        ,  1.        ,
                1.        ,  1.        ,  1.        ,  1.        ,  1.        ,
                1.        ,  1.        ,  1.        ,  1.        ,  1.        ,
                1.        ,  1.        ,  1.        ,  0.84986573,  0.4896242 ,
                0.13561368,  0.        ,  0.        ,  0.        ,  0.        ,
                0.        ,  0.        ,  0.        ,  0.        ,  0.        ], dtype=float32)
        >>> helixmask = SegmentExam().make_smooth_rectangular_mask(20, 20, 40, 0)
        >>> mask_row = helixmask.get_row(20)
        >>> EMNumPy.em2numpy(mask_row)
        array([ 0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  1.,  1.,  1.,
                1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,
                1.,  1.,  1.,  1.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.], dtype=float32)
        """
    
        if hel_height_pix is None:
            hel_height_pix = seg_size_pix
            
        falloff_line, rect_mask = self.generate_rectangular_mask_with_linear_falloffs(int(hel_width_pix),
        int(hel_height_pix), seg_size_pix, width_falloff)
        
        if width_falloff > 0:
            radial_quadrant = self.generate_radial_falloff_gradient(falloff_line)
        
            emmask = self.insert_radial_falloff_gradient_into_corners_of_rectangular_mask(radial_quadrant, falloff_line,
            rect_mask)
        else:
            emmask = EMNumPy.numpy2em(np.copy(rect_mask))
        
        resized_mask = self.resize_mask_to_segment_dimensions(emmask, seg_size_pix)
        
        return resized_mask
        
            
class SegmentExamWidth(SegmentExamMask):
    def find_local_extrema(self, fits, target='maxima', window=None):
        """
        Function from [SciPy-user] mailing list 'Finding local minima of greater than a given depth'

        #. Input: 1D/2D array of data and window size for minimum filter
        #. Output: ordered indices and minimum values
        #. Usage: ind, minima = find_local_extrema(array, target, window)
        
        """

        if window is None: 
            window=int(0.1*len(fits))

        fits = np.asarray(fits)
        if target == 'minima':
            from scipy.ndimage import minimum_filter
            minfits = minimum_filter(fits, size=window, mode='wrap')
        elif target == 'maxima':
            from scipy.ndimage import maximum_filter
            minfits = maximum_filter(fits, size=window, mode='wrap')

        minima_mask = fits == minfits
        good_indices = np.arange(len(fits))[minima_mask]
        good_fits = fits[minima_mask]
        order = good_fits.argsort()

        return good_indices[order], good_fits[order]

    def measure_peakdist(self, rowsaddimg=None, segsizepix=None, pixelsize=None):
        """
        * Function to measure distance between two symmetrical peaks of 1D helix width projection

        #. Input: rowsaddimg = projection to be measured, segsizepix = segment size (pixel), pixelsize
        #. Output: width of helix in Angstrom
        #. Usage: width = measure_peakdist(rowsaddimg, segsizepix, pixelsize)
        
        """
        from scipy import interpolate

        if rowsaddimg is None: rowsaddimg = self.rowsaddimg
        if segsizepix is None: segsizepix = self.segsizepix
        if pixelsize is None: pixelsize = self.pixelsize

        rowsaddimgnp = EMNumPy.em2numpy(rowsaddimg)
        rowsaddimgcp = np.copy(rowsaddimgnp)

        while True:
            try:
                img1dlineleft, img1dlineright = np.hsplit(rowsaddimgcp, 2)
                break
            except ValueError:
                rowsaddimgcp = np.append(rowsaddimgcp, max(rowsaddimgcp))

        img1dlineleft = np.flipud(img1dlineleft)

        cols = np.arange(0, len(img1dlineleft))
        colspol = np.arange(0, len(cols), 0.01)
        t = interpolate.splrep(cols, img1dlineleft, k=3, s=0)
        img1dlineleftpol = interpolate.splev(colspol, t)

        t = interpolate.splrep(cols, img1dlineright, k=3, s=0)
        img1dlinerightpol = interpolate.splev(colspol, t)

        ind, minima = self.find_local_extrema(img1dlineleftpol, target='minima')
        leftmin = colspol[ind[0]]
        # loop until value is significantly away from center of image
        it = iter(ind)
        while (leftmin < 0.2*self.helixwidthpix):
            try:
                leftmin = colspol[next(it)]
            except:
                break

        # sort according to real for display purposes
        ind, minima = self.find_local_extrema(img1dlinerightpol, target='minima')
        rightmin = colspol[ind[0]]
        it = iter(ind)
        while (rightmin < 0.2*self.helixwidthpix):
            try:
                rightmin = colspol[next(it)]
            except:
                break

#        self.log.dlog('min1: {0}, min2: {1}'.format(leftmin, rightmin))
        width = (rightmin + leftmin)*pixelsize

        return width
    

class SegmentExamLayerCorrelation(SegmentExamWidth):
    def compute_radii_for_fourier_mask(self, layer_line_region, segment_size, pixelsize):
        """
        >>> from spring.segment2d.segmentexam import SegmentExam
        >>> SegmentExam().compute_radii_for_fourier_mask(((0.1, 0.2)), 100, 2.5) 
        array([ 25.,  50.])
        """
        radii_pixel = np.array(layer_line_region) * pixelsize * segment_size
        radii_pixel = np.round(radii_pixel)
        
        return radii_pixel
        
    def generate_series_of_circular_masks_from_radii(self, radii_pixel, segment_size):
        masks = []
        for each_radius in np.arange(radii_pixel[0], radii_pixel[-1]):
            outer = model_circle(each_radius, segment_size, segment_size)
            inner = model_circle(each_radius - 1, segment_size, segment_size)
            
            ring = outer - inner
            masks.append(ring)
            
        return masks
    
            
    def compute_power_correlations_with_rings(self, avg_periodogram, masked_power, masks, segment_ids):
        power = EMData()
        correlations = []
        log_info = []
        for each_local_seg_id, each_stack_id in enumerate(segment_ids):
            power.read_image(masked_power, each_local_seg_id)
            
            correlation_of_one = [ccc(avg_periodogram, power, each_ring) for each_ring in masks]
            mean_ccc = np.mean(correlation_of_one)
            correlations.append(mean_ccc)
            log_info += [[each_stack_id, each_local_seg_id, mean_ccc]]

        msg = tabulate(log_info, ['stack_id', 'local_id', 'ccc'])

        self.log.ilog('The amplitudes of the following segments correlate with the avaraged power' + \
        'spectrum.\n{0}'.format(msg))
        
        return correlations
    
            
    def enter_correlation_values_in_database(self, correlations, segment_ids):
        session = SpringDataBase().setup_sqlite_db(base)
        
        for each_id, each_stack_id in enumerate(segment_ids):
            each_segment = session.query(SegmentTable).get(each_stack_id + 1)
            each_segment.ccc_layer = correlations[each_id]
            session.merge(each_segment)
            
        helices = session.query(HelixTable).all()
        for each_helix in helices:
            cc_per_helix = session.query(SegmentTable.ccc_layer).filter(SegmentTable.helix_id == HelixTable.id)
            avg_ccc_layer = np.mean([each_cc for each_cc in cc_per_helix])
            
            each_helix.avg_ccc_layer = avg_ccc_layer
            each_helix.ccc_layer_position_start = self.res_ccc_range[0]
            each_helix.ccc_layer_position_end = self.res_ccc_range[1]
            session.merge(each_helix)
            
        session.commit()
        

class SegmentExamVisualize(SegmentExamLayerCorrelation):
    def split_quarters(self, addpowimgenh=None):
        """
        * Function to split enhanced power spectrum (EMData object) into lower right quarter

        #. Input: avg_periodogram_enhanced = added power spectrum, segment size (pixel)
        #. Output: lower right quarter
        #. Usage: addpowimgenh1st = split_quarters(avg_periodogram_enhanced)
        
        """
        if addpowimgenh is None: addpowimgenh = self.avg_periodogram_enhanced
        segsizepix  = addpowimgenh.get_xsize()

        addpowimgenh1st = Util.window(addpowimgenh, segsizepix/2, segsizepix/2, 1, segsizepix/4, segsizepix/4, 0)

        return addpowimgenh1st
    

    def setup_fourxtwo(self, figno=None):
        """
        * Function to setup 4 x 2 subplot grid for diagnostic output

        #. Input: figno = figure number
        #. Output: figure
        #. Usage figure = setupfourxtwo(figno)
        
        """
        segmentexam_plot = DiagnosticPlot()
        if len(self.feature_set.parameters) < 15:
            self.fig = segmentexam_plot.add_header_and_footer(self.feature_set)
        else:    
            self.fig = segmentexam_plot.fig
        
        self.ax1 = segmentexam_plot.plt.subplot2grid((2,4), (0,0), colspan=1, rowspan=1)
        self.ax2 = segmentexam_plot.plt.subplot2grid((2,4), (1,0), colspan=1, rowspan=1)
        self.ax3 = segmentexam_plot.plt.subplot2grid((2,4), (0,1), colspan=1, rowspan=1)
        self.ax4 = segmentexam_plot.plt.subplot2grid((2,4), (1,1), colspan=1, rowspan=1)
        self.ax5 = segmentexam_plot.plt.subplot2grid((2,4), (0,2), colspan=1, rowspan=1)
        self.ax6 = segmentexam_plot.plt.subplot2grid((2,4), (1,2), colspan=1, rowspan=1)
        self.ax7 = segmentexam_plot.plt.subplot2grid((2,4), (0,3), colspan=1, rowspan=1)
        self.ax8 = segmentexam_plot.plt.subplot2grid((2,4), (1,3), colspan=1, rowspan=1)

        subplot_collection = [self.ax1, self.ax2, self.ax3, self.ax4, self.ax5, self.ax6, self.ax7, self.ax8]
        subplot_collection = segmentexam_plot.set_fontsize_to_all_ticklabels_of_subplots(subplot_collection)

        return self.fig


    def display_average_and_variance(self, twodavg=None, twodvar=None):
        """
        * Function to add average and variance images to diagnostic output plot

        #. Input: 2D average, 2d variance
        #. Output: subplot ax1, subplot ax3
        #. Usage: ax1, ax3 = display_average_and_variance(twodavg, twodvar)
        
        """
        if twodavg is None: twodavg = self.twodavg
        if twodvar is None: twodvar = self.twodvar

        twodavg = Micrograph().adjust_gray_values_for_print_and_optimal_display(twodavg)
        arrtwodavg = np.copy(EMNumPy.em2numpy(twodavg))

        # ax1: 2D average of width
        self.ax1.set_title('Average of segments', fontsize=8)
        self.ax1.set_xticks([])
        self.ax1.set_yticks([])
        self.ax1.imshow(arrtwodavg, cmap=cm.gray, interpolation='nearest')
        self.log.ilog('Avarage image included in montage')

#        twodvar = Micrograph().adjust_gray_values_for_print_and_optimal_display(twodvar)
        self.arrtwodvar = np.copy(EMNumPy.em2numpy(twodvar))
        # ax3 variance image
        self.ax3.set_title('Variance of segments', fontsize=8)
        self.ax3.set_xticks([])
        self.ax3.set_yticks([])
        self.ax3.imshow(self.arrtwodvar, cmap=cm.gray, interpolation='nearest')
        self.log.ilog('Variance image included in montage')

        return self.ax1, self.ax3
    

    def add_width_profile_from_avg_and_var(self, widthavg, widthvar, pixelsize, axx, quantity='width'):
        widthavgline = np.copy(EMNumPy.em2numpy(widthavg))
        widthvarline = np.copy(EMNumPy.em2numpy(widthvar))
        distance = self.measure_peakdist(widthavg, len(widthavgline), pixelsize)
        
        Angstrom = self.pixelsize * np.arange(len(widthavgline))
        table_data = [[each_ang, each_avg, each_var] for each_ang, each_avg, each_var in zip(Angstrom, widthavgline, widthvarline)]

        msg = tabulate(table_data, ['{0} (Angstrom)'.format(quantity.title()),'Intensity average', 'Intensity variance'])
        self.log.ilog('The following average/variance {0} profile was determined:\n{1}'.format(quantity, msg))
        
        axx.set_yticks([])
        axx.set_xlabel('Length (Angstrom)', fontsize=8)
        axx.set_ylabel('Image density', fontsize=8)
        axx.grid(True)
        axxtwin = axx.twinx()
        width_in_angstrom = (np.arange(len(widthavgline)) - len(widthavgline) / 2) * pixelsize
        axx.plot(width_in_angstrom, widthavgline, linewidth=.5, label='Average', color='r')
        axxtwin.plot(width_in_angstrom, widthvarline, linewidth=.5, label='Variance', color='b')
        axx.set_xlim(min(width_in_angstrom), max(width_in_angstrom))
        axx.set_ylim(min(widthavgline), 1.2 * max(widthavgline))
        axxtwin.set_ylim(min(widthvarline), 1.2 * max(widthvarline))
        axxtwin.set_yticks([])
        axx.text(0.5, max(widthavgline), 'Distance: %d A' % (distance), fontsize=5)
        axxtwin.legend(loc=3, prop=font_manager.FontProperties(size=5))
        axx.legend(loc=4, prop=font_manager.FontProperties(size=5))
        
        return axx


    def add_width_histogram_next_to_width_profile(self, widths):
        self.ax4.set_title('Helix width distribution', fontsize=8)
        nobins = self.helixwidth / 10
        self.ax4.hist(widths, nobins, facecolor='green', rwidth=1)
        self.ax4.set_xlim(0, 1.5 * self.helixwidth)
        self.ax4.set_ylabel('Number of segments', fontsize=8)
        self.ax4.set_xlabel('Width (Angstrom)', fontsize=8)
        self.ax4.grid(True)

    def visualize_widthprofile_and_histogram(self, widths=None, widthavg=None, widthvar=None, pixelsize=None):
        """
        * Function to add width profile to diagnostic output plot

        Input: widths = list of widths, widthavg = average of width, widthvar = variance of width, \
            pixelsize
        Output: subplot ax2, ax4
        Usage: ax2, ax4 = visualize_widthprofile_and_histogram(widths, widthavg, widthvar, pixelsize)
        
        """
        if widths is None: widths = self.widths
        if widthavg is None: widthavg = self.widthavg
        if widthvar is None: widthvar = self.widthvar
        if pixelsize is None: pixelsize = self.pixelsize

        self.ax2.set_title('Helix width profile', fontsize=8)
        self.ax2 = self.add_width_profile_from_avg_and_var(widthavg, widthvar, pixelsize, self.ax2)
        self.log.ilog('Width profile of average segment included in montage')
        self.add_width_histogram_next_to_width_profile(widths)

        return self.ax2, self.ax4

    def display_power_spectra_enhanced_and_collapsed(self, avg_periodogram=None, avg_periodogram_enhanced=None,
    avg_collapsed_power_line=None, avg_collapsed_line_enhanced=None):
        """
        * Function to visualize power spectra: sum of power spectra, enhanced sum and their collapsed 1D profile

        #. Input: avg_periodogram = sum of power spectra (img), avg_periodogram_enhanced = enhanced sum \
            of power spectra, avg_collapsed_power_line = collapsed profile of power spectrum (img), \
            avg_collapsed_line_enhanced = collapsed profile of enhanced power spectrum
        #. Output: subplots ax5, ax6, ax7, ax8
        #. Usage: ax5, ax6, ax7, ax8 = display_power_spectra_enhanced_and_collapsed(avg_periodogram, 
            avg_periodogram_enhanced, avg_collapsed_power_line, avg_collapsed_line_enhanced)
        
        """

        if avg_periodogram is None: 
            avg_periodogram = self.avg_periodogram
        if avg_periodogram_enhanced is None: 
            avg_periodogram_enhanced = self.avg_periodogram_enhanced
        if avg_collapsed_power_line is None: 
            avg_collapsed_power_line = self.avg_collapsed_power_line
        if avg_collapsed_line_enhanced is None: 
            avg_collapsed_line_enhanced = self.avg_collapsed_line_enhanced

        # ax5: 2D avaraged power spectrum
        self.ax5.set_title('Averaged powerspectrum', fontsize=8)

        def fourier_pix(x, pos):
            resol = abs(int(x - avg_periodogram.get_xsize() / 2))
            return resol

        formatter_x = FuncFormatter(fourier_pix)
        self.ax5.xaxis.set_major_formatter(formatter_x)
        self.ax5.yaxis.set_major_formatter(formatter_x)

        self.ax5.set_xlabel('Fourier pixel', fontsize=8)

        avg_periodogram = Micrograph().adjust_gray_values_for_print_and_optimal_display(avg_periodogram)
        self.arraddpowimg = np.copy(EMNumPy.em2numpy(avg_periodogram))
        img = self.ax5.imshow(self.arraddpowimg, cmap=cm.hot, interpolation='nearest')

        # colorbar
        cax = self.fig.add_axes([0.93, 0.6, 0.01, 0.25])
        cbar = self.fig.colorbar(img, cax)
        for t in cbar.ax.get_yticklabels():
            t.set_fontsize(5)

        self.log.ilog('Averaged power spectrum added')

        # ax7: 2D avaraged enhanced power spectrum
        self.ax7.set_title('Enhanced \n(B-factor compensated)', fontsize=8)

        def res(x, pos):
            resol = '{0:.3f}'.format(
                    self.nyquist_frequency * (x - avg_periodogram.get_xsize() / 2)  / float(avg_periodogram.get_xsize() / 2)) 
            return resol

        formatter_x = FuncFormatter(res)
        self.ax7.xaxis.set_major_formatter(formatter_x)
        self.ax7.yaxis.set_major_formatter(formatter_x)
        self.ax7.set_xlabel('Resolution (1/Angstrom)', fontsize=8)

        avg_periodogram_enhanced = \
        Micrograph().adjust_gray_values_for_print_and_optimal_display(avg_periodogram_enhanced)
        
        self.arraddpowimgenh = np.copy(EMNumPy.em2numpy(avg_periodogram_enhanced))
        self.ax7.imshow(self.arraddpowimgenh, cmap=cm.hot, interpolation='nearest')
        self.log.ilog('Averaged enhanced power spectrum added')

        addpowimgenh1st = self.split_quarters(avg_periodogram_enhanced)
        # ax6: first quarter
        self.ax6.set_title('Lower right quadrant', fontsize=8)
        
        self.arraddcolline = np.copy(EMNumPy.em2numpy(avg_collapsed_power_line))
        self.arraddcollineenh = np.copy(EMNumPy.em2numpy(avg_collapsed_line_enhanced))
        self.make_oneoverres(self.arraddcolline)

        def res(x, pos):
            resol = '{0:.3f}'.format(self.nyquist_frequency * x / float(len(self.arraddcolline))) 
            return resol

        formatter_x = FuncFormatter(res)
        self.ax6.xaxis.set_major_formatter(formatter_x)
        self.ax6.yaxis.set_major_formatter(formatter_x)
        self.ax6.set_xlabel('Resolution (1/Angstrom)', fontsize=8)

        self.arraddpowimgenh1st = np.copy(EMNumPy.em2numpy(addpowimgenh1st))
        self.ax6.imshow(self.arraddpowimgenh1st, cmap=cm.hot, interpolation='nearest')
        self.log.ilog('Upper left quadrant added.')

        # ax8: 1D collapsed power spectrum
        self.ax8.set_title('Collapsed powerspectrum', fontsize=8)

        self.ax8twin = self.ax8.twinx()
        self.ax8twin.plot(self.arrresolution, self.arraddcolline, linewidth = .5, color='b', label='collapsed')
        self.ax8.plot(self.arrresolution, self.arraddcollineenh, linewidth = .5, color='r', label='enhanced')
        self.ax8twin.set_yticks([])
        self.ax8.legend(loc=1, prop=font_manager.FontProperties(size='x-small'))
        self.ax8.set_xlabel('Resolution (1/Angstrom)', fontsize=8)
        self.ax8.set_yticks([])
        self.ax8.set_xlim(0, max(self.arrresolution))
        self.ax8twin.set_xlim(0, max(self.arrresolution))
        self.ax8.grid(True)

        table_data = [[each_res, each_col, each_colen]
                       for each_res, each_col, each_colen in zip(self.arrresolution, self.arraddcolline, self.arraddcollineenh)]
        
        msg = tabulate(table_data, ['Resolution (1/Angstrom)', 'Intensity', 'Intensity (enhanced)'])
        self.log.ilog('The following collapsed power spectrum profile was determined:\n{0}'.format(msg))

        return self.ax5, self.ax6, self.ax7, self.ax8

    def make_oneoverres(self, arr=None, pixelsize=None):
        """
        * Function to generate an array of resolution in reciprocal Angstrom

        #. Input: array, pixelsize
        #. Output: array of reciprocal resolution (1/Angstrom)
        #. Usage: arrresolution = make_overoverres(arr, pixelsize)
        
        >>> from spring.segment2d.segmentexam import SegmentExam
        >>> SegmentExam().make_oneoverres(range(10), 10)
        array([ 0.        ,  0.00555556,  0.01111111,  0.01666667,  0.02222222,
                0.02777778,  0.03333333,  0.03888889,  0.04444444,  0.05      ])
        >>> SegmentExam().make_oneoverres(range(25), 1)
        array([ 0.        ,  0.02083333,  0.04166667,  0.0625    ,  0.08333333,
                0.10416667,  0.125     ,  0.14583333,  0.16666667,  0.1875    ,
                0.20833333,  0.22916667,  0.25      ,  0.27083333,  0.29166667,
                0.3125    ,  0.33333333,  0.35416667,  0.375     ,  0.39583333,
                0.41666667,  0.4375    ,  0.45833333,  0.47916667,  0.5       ])
        >>> 1/SegmentExam().make_oneoverres(range(25), 1)
        array([         inf,  48.        ,  24.        ,  16.        ,
                12.        ,   9.6       ,   8.        ,   6.85714286,
                 6.        ,   5.33333333,   4.8       ,   4.36363636,
                 4.        ,   3.69230769,   3.42857143,   3.2       ,
                 3.        ,   2.82352941,   2.66666667,   2.52631579,
                 2.4       ,   2.28571429,   2.18181818,   2.08695652,   2.        ])
        """
        if arr is None: arr = self.arraddcolline
        if pixelsize is None: pixelsize = self.pixelsize

        self.nyquist_frequency = 1/(float(pixelsize)*2)
        self.arrresolution = np.linspace(0, self.nyquist_frequency, len(arr))

        return self.arrresolution

    def cleanup(self, *files):
        """
        * Function to clean up intermediate image files

        Input: arbitrary number of files
        Ouput: None
        Usage: cleanup(thisfile, anotherfile)
        """
        for rmfile in files:
            if os.path.isfile(rmfile):
                os.remove(rmfile)
                self.log.ilog('Intermediate file {0} was removed'.format(rmfile))

class SegmentExam(SegmentExamVisualize):


    def collapse_periodograms(self, avg_periodogram, avg_periodogram_enhanced):
        avg_collapsed_power_line = self.collapse_power(avg_periodogram)
        
        avg_collapsed_line_enhanced = self.collapse_power(avg_periodogram_enhanced)
        
        return avg_collapsed_power_line, avg_collapsed_line_enhanced
    

    def write_avg_periodograms(self, avg_periodogram, power_img, power_enhanced_img):
        avg_periodogram.write_image(power_img)
        avg_periodogram_enhanced = self.enhance_power(avg_periodogram, self.pixelsize)
        avg_periodogram_enhanced.write_image(power_enhanced_img)
            
        return avg_periodogram_enhanced
    

    def add_power_spectra_from_verticalized_stack(self, infilestack, segment_ids, helixwidth=None,
    masked_infilestack=None, power_infilestack=None, padsize=4):
        """
        * Function to compute sum of in-planed rotated segments 
        """
        self.log.fcttolog()
        self.log.in_progress_log()
        segment = EMData()
        segment.read_image(infilestack)
        segment_size = segment.get_xsize()
        avg_periodogram = model_blank(padsize * segment_size, padsize * segment_size, 1, 0)

        if helixwidth is None:
            helixwidth = segment_size
        helixmask = self.make_smooth_rectangular_mask(helixwidth, segment_size * 0.6, segment_size, 0.15)
        log_info = 'The following segments are fourier-transformed and their amplitudes averaged.' + \
        '\n(Local_id, Segment_id)\n'
        for each_local_seg_id, each_seg_id in enumerate(segment_ids):
            segment.read_image(infilestack, each_local_seg_id)
            segment *= helixmask
            if masked_infilestack is not None:
                segment.write_image(masked_infilestack, each_local_seg_id)
            log_info += '({0}, {1}) '.format(each_local_seg_id, each_seg_id)
            if padsize > 1:
                segment = Util.pad(segment, padsize * segment_size, padsize * segment_size, 1, 0, 0, 0, '0')
            ps = periodogram(segment)
            avg_periodogram += ps
            if power_infilestack is not None:
                ps.write_image(power_infilestack, each_local_seg_id)
        self.log.ilog(log_info)

        return avg_periodogram
    

    def compute_avg_and_var_of_width_and_image(self, infilestack, temp_rowsadd):
        widthavg, widthvar = add_series(temp_rowsadd)
        twodavg, twodvar = add_series(infilestack)
        os.remove(temp_rowsadd)
        
        return widthavg, widthvar, twodavg, twodvar
    

    def determine_width(self, infilestack, segsizepix, segment_ids):
        """
        * Function to project width profile of segments

        #. Input: stackfile, segment size (pixel)
        #. Output: width average profile, width variance profile, measured width list
        #. Usage: widthavg, widthvar, widths = determine_width(infilestack, segsizepix)
        """

        self.log.fcttolog()
        widths = []
        
        segment = EMData()
        temp_rowsadd = os.path.join(self.tempdir,'rowsaddimg.hdf')
        
        log_info = []
        for each_local_seg_id, each_stack_id in enumerate(segment_ids):
            segment.read_image(infilestack, each_local_seg_id)
            rowsaddimg = self.project_helix(segment)
            rowsaddimg.write_image(temp_rowsadd, each_local_seg_id)

            width = self.measure_peakdist(rowsaddimg, self.segsizepix, self.pixelsize)
            widths.append(width)
            log_info += [[each_stack_id, each_local_seg_id, width]]

        msg = tabulate(log_info, ['stack_id', 'local_id', 'width (Angstrom)'])
        self.log.ilog('The following segments were projected and their width is measured:\n{0}'.format(msg))

        return temp_rowsadd, widths
    

    def correlate_layer_lines_of_average_power_with_individual_segments(self, avg_periodogram, masked_power,
    segment_ids):
        segment = EMData()
        segment.read_image(masked_power)
        segment_size = segment.get_xsize()
        
        if self.power_reference:
            ref_periodogram = EMData()
            ref_periodogram.read_image(self.power_input)
            
            if avg_periodogram.get_xsize() != ref_periodogram.get_xsize():
                msg = 'The provided reference power spectrum does not have the same size as the power spectra of ' + \
                'the segments ({0} vs. {1} '.format(ref_periodogram.get_xsize(), avg_periodogram.get_xsize()) + \
                'pixels). Double-check the origin of the reference power spectrum. Was it generated using ' + \
                '{0} with the same processing options, e.g. \'Power spectrum '.format(self.feature_set.progname) + \
                'resolution cutoff in 1/Angstrom\'?'
                raise ValueError(msg)
            cc_periodogram = ref_periodogram
        else:
            cc_periodogram = avg_periodogram

        fourier_radii = self.compute_radii_for_fourier_mask(self.res_ccc_range, segment_size, self.pixelsize)
        masks = self.generate_series_of_circular_masks_from_radii(fourier_radii, segment_size)
        correlations = self.compute_power_correlations_with_rings(cc_periodogram, masked_power, masks, segment_ids)
        
        return correlations
        

    def visualize_power_avg_and_width_analysis(self, widthavg, widthvar, widths, twodavg, twodvar, avg_periodogram,
    avg_periodogram_enhanced, avg_collapsed_power_line, avg_collapsed_line_enhanced):
        """
        * Function to combine output of width and power spectra analysis into single summary sheet

        """
            
        self.log.fcttolog()

        self.setup_fourxtwo()

        self.display_average_and_variance(twodavg, twodvar)
        self.visualize_widthprofile_and_histogram(widths, widthavg, widthvar, self.pixelsize)
        
        self.display_power_spectra_enhanced_and_collapsed(avg_periodogram, avg_periodogram_enhanced,
        avg_collapsed_power_line, avg_collapsed_line_enhanced)

        self.fig.savefig(self.outfile)
        
        return self.fig
    

    def copy_database_and_filter_segment_ids(self):
        shutil.copy(self.spring_path, 'spring.db')

        self.curvature_range, self.ccc_layer_range = SegmentSelect().convert_curvature_ccc_layer_range('spring.db',
        self.straightness_selection, self.curvature_range_perc, self.ccc_layer_selection, self.ccc_layer_range_perc)
        
        segment_ids, excluded_segment_counts = \
        SegmentSelect().filter_non_orientation_parameters_based_on_selection_criteria(self)

        return segment_ids


    def add_up_power_spectra(self):
        OpenMpi().setup_and_start_mpi_version_if_demanded(self.mpi_option, self.feature_set, self.cpu_count)
        self.tempdir = Temporary().mktmpdir(self.temppath)
        
        self.infilestack, self.segsizepix, self.helixwidthpix, self.pixelsize = \
        self.apply_binfactor(self.binfactor, self.infilestack, self.segsizepix, self.helixwidthpix, self.pixelsize)
        
        self.log.plog(10)

        segment_ids = self.copy_database_and_filter_segment_ids()

        masked_infilestack = os.path.join(self.tempdir, 'infilestack-masked.hdf')
        power_infilestack = os.path.join(self.tempdir, 'infilestack-power.hdf')
        
        avg_periodogram = self.add_power_spectra_from_verticalized_stack(self.infilestack, segment_ids,
        self.helixwidthpix, masked_infilestack, power_infilestack)
        
        self.log.plog(40)
        avg_periodogram_enhanced = self.write_avg_periodograms(avg_periodogram, self.power_img, self.power_enhanced_img)
        
        avg_collapsed_power_line, avg_collapsed_line_enhanced = self.collapse_periodograms(avg_periodogram,
        avg_periodogram_enhanced)
        
        self.log.plog(60)
                     
        if self.layer_ccc_option:
            correlations = self.correlate_layer_lines_of_average_power_with_individual_segments(avg_periodogram,
            power_infilestack, segment_ids)
            
            self.enter_correlation_values_in_database(correlations, segment_ids)
        os.remove(power_infilestack)
            
        temp_rowsadd, self.widths = self.determine_width(masked_infilestack, self.segsizepix, segment_ids)
        
        widthavg, widthvar, twodavg, twodvar = self.compute_avg_and_var_of_width_and_image(masked_infilestack,
        temp_rowsadd)
        
        self.log.plog(80)
        
        os.remove(masked_infilestack)
        
        self.visualize_power_avg_and_width_analysis(widthavg, widthvar, self.widths, twodavg, twodvar, avg_periodogram,
        avg_periodogram_enhanced, avg_collapsed_power_line, avg_collapsed_line_enhanced)
        
        self.log.plog(90)
        
        self.cleanup(self.infilestack)
            
        os.rmdir(self.tempdir)
        self.log.endlog(self.feature_set)
        
        
def main():
    # Option handling
    parset = SegmentExamPar()
    mergeparset = OptHandler(parset)
    ######## Program
    stack = SegmentExam(mergeparset)
    stack.add_up_power_spectra()

if __name__ == '__main__':
    main()



# Author: Carsten Sachse 1-Jan-2012
# Copyright: EMBL (2010 - 2018)
# License: see license.txt for details

from EMAN2 import EMData
from spring.csinfrastr.csproductivity import OpenMpi
from spring.segment2d.segmentexam import SegmentExam, SegmentExamPar
import os


class SegmentExamMpi(SegmentExam):

    def generate_local_name_for_reduction(self, emdata_file, rank):
        """
        >>> from spring.segment2d.segmentexam_mpi import SegmentExamMpi
        >>> SegmentExamMpi().generate_local_name_for_reduction('ps_234567891.hdf', 2)
        'ps_2345678912.hdf'
        """
        local_emdata_file = '{pre}{rank}{ext}'.format(pre=os.path.splitext(emdata_file)[0], rank=rank,
        ext=os.path.splitext(emdata_file)[1])
        
        return local_emdata_file
    

    def reduce_emdata_from_memory_on_main_node(self, widthavg):
        widthavg_file = 'width_avg.hdf'
        local_width_avg = self.generate_local_name_for_reduction(widthavg_file, self.rank)
        widthavg.write_image(local_width_avg)
        local_width_avg = self.comm.gather(local_width_avg, root=0)
        if self.rank == 0:
            widthavg = OpenMpi().reduce_emdata_on_main_node(widthavg, local_width_avg)
        else:
            widthavg = None
            
        return widthavg


    def prepare_segmentexam_mpi(self):
        self.comm, self.rank, self.size, self.log, self.tempdir = OpenMpi().setup_mpi_and_simultaneous_logging(self.log,
        self.feature_set.logfile, self.temppath)
        
        if self.rank == 0:
            segment_ids = self.copy_database_and_filter_segment_ids()
            segment_ids = OpenMpi().split_sequence_evenly(segment_ids, self.size)
        else:
            segment_ids = None
        segment_ids = self.comm.scatter(segment_ids, root=0)

        binned_stack = self.tempdir + os.path.basename(self.infilestack)
            
        self.infilestack, self.segsizepix, self.helixwidthpix, self.pixelsize = self.apply_binfactor(self.binfactor,
        self.infilestack, self.segsizepix, self.helixwidthpix, self.pixelsize, segment_ids, binned_stack)

        self.segsizepix = self.comm.bcast(self.segsizepix, root=0)
        self.helixwidthpix = self.comm.bcast(self.helixwidthpix, root=0)
        self.pixelsize = self.comm.bcast(self.pixelsize, root=0)
        self.comm.barrier()
        
        if self.rank == 0:
            self.log.plog(10)
            
        return segment_ids
    

    def add_powers_locally_and_reduce_on_main_node(self, segment_ids):
        masked_infilestack = os.path.join(self.tempdir, 'infilestack-masked.hdf')
        power_infilestack = os.path.join(self.tempdir, 'infilestack-power.hdf')
        
        avg_periodogram = self.add_power_spectra_from_verticalized_stack(self.infilestack, segment_ids,
        self.helixwidthpix, masked_infilestack, power_infilestack)
        
        local_power_img = self.generate_local_name_for_reduction(self.power_img, self.rank)
        if self.enhanced_power_option:
            local_power_enhanced_img = self.generate_local_name_for_reduction(self.power_enhanced_img, self.rank)
        else:
            local_power_enhanced_img = self.generate_local_name_for_reduction(os.path.splitext(self.power_img)[0] + \
            '_enh' + os.path.splitext(self.power_img)[-1], self.rank)
        avg_periodogram_enhanced = self.write_avg_periodograms(avg_periodogram, local_power_img,
        local_power_enhanced_img)
        
        self.comm.barrier()
        local_power_img = self.comm.gather(local_power_img, root=0)
        local_power_enhanced_img = self.comm.gather(local_power_enhanced_img, root=0)
        if self.rank == 0:
            avg_periodogram = OpenMpi().reduce_emdata_on_main_node(avg_periodogram, local_power_img)
            avg_periodogram.write_image(self.power_img)
            
            avg_periodogram_enhanced = OpenMpi().reduce_emdata_on_main_node(avg_periodogram_enhanced,
            local_power_enhanced_img)
            
            if self.enhanced_power_option:
                avg_periodogram_enhanced.write_image(self.power_enhanced_img)
            avg_collapsed_power_line, avg_collapsed_line_enhanced = self.collapse_periodograms(avg_periodogram,
            avg_periodogram_enhanced)
            
            self.log.plog(60)
        else:
            avg_collapsed_power_line = None
            avg_collapsed_line_enhanced = None
            
        self.comm.barrier()
        
        return avg_periodogram, power_infilestack, masked_infilestack, avg_periodogram_enhanced, \
        avg_collapsed_power_line, avg_collapsed_line_enhanced
    

    def correlate_layer_line_region_mpi(self, segment_ids, avg_periodogram, power_infilestack):
        if self.layer_ccc_option:
            avg_periodogram = EMData()
            avg_periodogram.read_image(self.power_img)

            correlations = self.correlate_layer_lines_of_average_power_with_individual_segments(avg_periodogram,
            power_infilestack, segment_ids)
            
            correlations = self.comm.gather(correlations, root=0)
            segment_ids = self.comm.gather(segment_ids, root=0)
            if self.rank == 0:
                correlations = OpenMpi().merge_sequence_of_sequences(correlations)
                segment_ids = OpenMpi().merge_sequence_of_sequences(segment_ids)
                self.enter_correlation_values_in_database(correlations, segment_ids)
        self.comm.barrier()
        os.remove(power_infilestack)


    def determine_width_from_collapsed_profile_mpi(self, segment_ids, masked_infilestack):
        temp_rowsadd, widths = self.determine_width(masked_infilestack, self.segsizepix, segment_ids)
        common_masked_stack = 'common_masked.hdf'
        OpenMpi().gather_stacks_from_cpus_to_common_stack(self.comm, masked_infilestack, common_masked_stack)
        common_rows = 'common_rows.hdf'
        OpenMpi().gather_stacks_from_cpus_to_common_stack(self.comm, temp_rowsadd, common_rows)
        self.widths = self.comm.gather(widths, root=0)
        
        return common_masked_stack, common_rows


    def visualize_avg_var_widths_and_power_spectra_mpi(self, avg_periodogram, avg_periodogram_enhanced,
    avg_collapsed_power_line, avg_collapsed_line_enhanced, common_masked_stack, common_rows):
        if self.rank == 0:
            widthavg, widthvar, twodavg, twodvar = self.compute_avg_and_var_of_width_and_image(common_masked_stack,
            common_rows)
            
            os.remove(common_masked_stack)
            self.widths = OpenMpi().merge_sequence_of_sequences(self.widths)
            
            self.visualize_power_avg_and_width_analysis(widthavg, widthvar, self.widths, twodavg, twodvar,
            avg_periodogram, avg_periodogram_enhanced, avg_collapsed_power_line, avg_collapsed_line_enhanced)
            
            self.log.plog(80)

        self.cleanup(self.infilestack)

    def add_up_power_spectra(self):
        segment_ids  = self.prepare_segmentexam_mpi()
        
        avg_periodogram, power_infilestack, masked_infilestack, avg_periodogram_enhanced, avg_collapsed_power_line, \
        avg_collapsed_line_enhanced = self.add_powers_locally_and_reduce_on_main_node(segment_ids)
                     
        self.correlate_layer_line_region_mpi(segment_ids, avg_periodogram, power_infilestack)
        common_masked_stack, common_rows = self.determine_width_from_collapsed_profile_mpi(segment_ids, masked_infilestack)
        
        self.visualize_avg_var_widths_and_power_spectra_mpi(avg_periodogram, avg_periodogram_enhanced,
        avg_collapsed_power_line, avg_collapsed_line_enhanced, common_masked_stack, common_rows)
            
        os.rmdir(self.tempdir)
        self.comm.barrier()
        if self.rank == 0:
            self.log.endlog(self.feature_set)

def main():
    parset = SegmentExamPar()
    reduced_parset = OpenMpi().start_main_mpi(parset)
    
    ####### Program
    stack = SegmentExamMpi(reduced_parset)
    stack.add_up_power_spectra()

if __name__ == '__main__':
    main()



# Author: Carsten Sachse 
# Copyright: EMBL (2010 - 2018)
# License: see license.txt for details
"""
Program to plot parameters from segmented helices
"""
from EMAN2 import EMData, EMNumPy
from PyQt4.QtCore import SIGNAL, Qt
from PyQt4.QtGui import QApplication, QWidget, QStackedWidget, QSplitter, QComboBox
from collections import namedtuple
from matplotlib.font_manager import FontProperties
from spring.csinfrastr.csdatabase import SpringDataBase, base, SegmentTable, HelixTable, CtfMicrographTable
from spring.csinfrastr.csfeatures import Features
from spring.csinfrastr.cslogger import Logger
from spring.csinfrastr.csproductivity import DiagnosticPlot
from spring.csinfrastr.csreadinput import OptHandler
from spring.segment2d.segmentselect import SegmentSelect
from spring.springgui.springdataexplore import SpringCommon, SpringDataExplore, SpringDataExploreTablePane
from tabulate import tabulate
import numpy as np
import os
import shutil
import sys
from utilities import image_decimate

class SegmentPlotPar(object):
    """
    Class to initiate default dictionary with input parameters including help and range values and 
    status dictionary
    """
    def __init__(self):
        # package/program identity
        self.package = 'emspring'
        self.progname = 'segmentplot'
        self.proginfo = __doc__
        self.code_files = [self.progname]

        self.segmentplot_features = Features()
        self.feature_set = self.segmentplot_features.setup(self)
        
        self.define_parameters_and_their_properties()
        self.define_program_states()
    
    def define_parameters_and_their_properties(self):
        self.feature_set = self.set_inp_spring_path(self.feature_set)
        self.feature_set = self.segmentplot_features.set_interactive_vs_batch_mode(self.feature_set)
        
        self.feature_set = self.segmentplot_features.set_diagnostic_prefix(self.feature_set, 'intermediate', 
        'Batch mode')
        
        self.feature_set = self.set_quantities_to_be_plotted(self.feature_set)
        self.feature_set = self.set_set_size(self.feature_set)
        
        self.feature_set = self.segmentplot_features.set_selection_criteria_from_segment_table(self.feature_set)
        

    def define_program_states(self):

        self.feature_set.program_states['get_quantities_per_set']='Extract desired quantities from spring database'
        self.feature_set.program_states['plot_data_on_figure']='Prepare figures with desired quantities'


    def set_inp_spring_path(self, feature_set):
        inp3 = 'spring.db file'
        feature_set.parameters[inp3]='spring.db'
        feature_set.hints[inp3]='Requires spring.db from segment to plot helix parameters.'
        feature_set.properties[inp3]=feature_set.file_properties(1,['db'],'getFile')
        feature_set.level[inp3]='beginner'
        
        return feature_set
    
    
    def set_quantities_to_be_plotted(self, feature_set):
        inp3 = 'Quantities'
        feature_set.parameters[inp3]=str('coordinates')
        
        feature_set.properties[inp3]=feature_set.choice_properties(2, ['coordinates', 'in-plane_rotation', 'curvature',
        'defocus', 'astigmatism', 'layer-line correlation', 'classes', 'class_models'], 'QComboBox')
        
        feature_set.hints[inp3]=SpringDataBase().get_hints_from_segment_table(feature_set.properties[inp3].choices)  
        feature_set.level[inp3]='beginner'
        
        return feature_set
    
    
    def set_interactive_vs_batch_mode(self, feature_set):
        inp3 = 'Interactive vs. batch mode'
        feature_set.parameters[inp3]='interactive'
        feature_set.hints[inp3]='Choose between interactive (enables quantity per helix view) vs. batch plotting mode.'
        feature_set.properties[inp3]=feature_set.choice_properties(2, ['interactive', 'batch'], 'QComboBox')
        feature_set.level[inp3]='expert'
        
        return feature_set
    
    
    def set_set_size(self, feature_set):
        inp3 = 'Set size'
        feature_set.parameters[inp3]='helix'
        feature_set.hints[inp3]='Choose set size to plot: chosen quantity per \'helix\', \'micrograph\' or \'data_set\'.'
        feature_set.properties[inp3]=feature_set.choice_properties(2, ['helix', 'micrograph', 'data_set'], 'QComboBox')
        feature_set.level[inp3]='expert'
        
        return feature_set
        
    
class SegmentPlotPrepare(object):
    """
    * Class that holds functions for examining segments from micrographs

    * __init__ Function to interpret multi-input parameters

    """
    def __init__(self, parset = None):
        self.log = Logger()
        if parset is not None:
            self.feature_set = parset
            p = self.feature_set.parameters

            self.infile = p['spring.db file']
            self.outfile_prefix = p['Diagnostic plot prefix']
            
            self.quantities = p['Quantities']
            self.batch_mode = p['Batch mode']
            self.set_size = p['Set size']

            self = SegmentSelect().define_selection_parameters_from_segment_table(self, p)



    def check_class_values_and_raise_error(self, quantities, y_plot):
        if y_plot[0] is None:
            msg = '{0} entry does not exist in specified spring.db. '.format(quantities) + \
            'Re-run segmentclass including database option to save class information.'
            raise ValueError(msg)


    def check_class_model_values_and_raise_error(self, quantities, y_plot):
        if y_plot[0] is None:
            msg = '{0} entry does not exist in specified spring.db. '.format(quantities) + \
            'Re-run segmentrefine3d or segclassmodel including database option to save class model information.'
            raise ValueError(msg)


    def check_ctf_values_and_raise_error(self, quantities, y_plot):
        if y_plot[0] is None:
            msg = '{0} entry does not exist in specified spring.db. '.format(quantities) + \
            'Re-run micctfdetermine or segment to save CTF-information.'
            raise ValueError(msg)


    def check_layerline_values_and_raise_error(self, quantities, y_plot):
        if y_plot[0] is None:
            msg = '{0} entry does not exist in specified spring.db. '.format(quantities)
            'Re-run segmentexam with the \'Compute layer-line correlation option\'.'
            raise ValueError(msg)


    def check_segmenting_values_and_raise_error(self, quantities, y_plot):
        if y_plot[0] is None:
            msg = '{0} entry does not exist in specified spring.db. Re-run segment to save relevant information.'.\
            format(quantities)
            raise ValueError(msg)


    def get_subset_if_selected(self, plot_sets, attr_selection, attr_in_or_exclude, attr_entries, attr_id):
        if attr_selection:
            attr_list = SegmentSelect().prepare_list_from_comma_separated_input(attr_entries, 'helix')
            if attr_in_or_exclude == 'include':
                plot_sets = [each_set for each_set in plot_sets if getattr(each_set, attr_id) in attr_list]
            elif attr_in_or_exclude == 'exclude':
                plot_sets = [each_set for each_set in plot_sets if getattr(each_set, attr_id) not in attr_list]

        return plot_sets


    def get_plot_set(self, set_size, obj):
        session = SpringDataBase().setup_sqlite_db(base)
        if set_size == 'helix':
            plot_sets = session.query(HelixTable).order_by(HelixTable.id).all()

            plot_sets = self.get_subset_if_selected(plot_sets, obj.mics_selection, obj.mics_in_or_exclude,
            obj.mics_entries, 'mic_id')

            plot_sets = self.get_subset_if_selected(plot_sets, obj.helices_selection, obj.helices_in_or_exclude,
            obj.helices_entries, 'id')
        elif set_size == 'micrograph':
            plot_sets = session.query(CtfMicrographTable).order_by(CtfMicrographTable.id).all()
            if obj.helices_selection:
                helix_sets = session.query(HelixTable).order_by(HelixTable.id).all()
                helix_sets = self.get_subset_if_selected(helix_sets, obj.helices_selection, obj.helices_in_or_exclude,
                obj.helices_entries, 'id')
                mic_ids = list(set([each_helix.mic_id for each_helix in helix_sets]))

                plot_sets = [each_set for each_set in plot_sets if each_set.id in mic_ids]
            obj.pixelsize = plot_sets[0].pixelsize

            plot_sets = self.get_subset_if_selected(plot_sets, obj.mics_selection, obj.mics_in_or_exclude,
            obj.mics_entries, 'id')
        elif set_size == 'data_set':
            plotset_tuple = namedtuple('plot_set', 'id')
            plot_sets = [plotset_tuple('entire')]
            
        return session, plot_sets
    

    def filter_segments_by_helix_and_micrograph_criteria(self, session, obj):
        segments = session.query(SegmentTable).order_by(SegmentTable.stack_id).all()

        included_segments_mics, excluded_mics_count = SegmentSelect().filter_segments_by_entry_string(segments,
        'mic_id', obj.mics_selection, obj.mics_in_or_exclude, obj.mics_entries, 'micrograph')
         
        included_segments_helices, excluded_helix_count = SegmentSelect().filter_segments_by_entry_string(segments,
        'helix_id', obj.helices_selection, obj.helices_in_or_exclude, obj.helices_entries, 'helix')
         
        combined_included_segments = list(set(included_segments_mics).intersection(included_segments_helices))
 
        return combined_included_segments
    

    def prepare_helix_and_coordinates(self, set_size, quantities):
        labels = ['Stack_id']
        if set_size == 'helix' and quantities != 'coordinates':
            labels += SpringDataBase().get_labels_from_table(SegmentTable, 'distance_from_start_A')
        elif quantities == 'coordinates':
            labels += SpringDataBase().get_labels_from_table(SegmentTable, 'x_coordinate_A', 'y_coordinate_A',
            'picked_x_coordinate_A', 'picked_y_coordinate_A')
            
        return labels


    def get_labels_according_to_quantities(self, set_size, quantities):
        """
        >>> from spring.segment2d.segmentplot import SegmentPlot
        >>> SegmentPlot().get_labels_according_to_quantities('helix', 'coordinates') #doctest: +NORMALIZE_WHITESPACE
        ['Stack_id', 'X-Coordinate (Angstrom)', 'Y-Coordinate (Angstrom)', 'Picked X-Coordinate (Angstrom)', 
        'Picked Y-Coordinate (Angstrom)']
        >>> SegmentPlot().get_labels_according_to_quantities('helix', 'curvature') 
        ['Stack_id', 'Distance Along Helix (Angstrom)', 'Curvature', 'Selected Curvature']
        >>> SegmentPlot().get_labels_according_to_quantities('data_set', 'curvature') 
        ['Stack_id', 'Curvature', 'Selected Curvature']
        >>> SegmentPlot().get_labels_according_to_quantities('micrograph', 'astigmatism') 
        ['Stack_id', 'Astigmatism (Angstrom)', 'Selected Astigmatism (Angstrom)']
        """
        labels = self.prepare_helix_and_coordinates(set_size, quantities)
            
        if quantities == 'in-plane_rotation':
            labels += SpringDataBase().get_labels_from_table(SegmentTable, 'inplane_angle')
            labels += ['Selected ' + SpringDataBase().get_labels_from_table(SegmentTable, 'inplane_angle')[0]]
        elif quantities in ['curvature']:
            labels += SpringDataBase().get_labels_from_table(SegmentTable, 'curvature')
            labels += ['Selected ' + SpringDataBase().get_labels_from_table(SegmentTable, 'curvature')[0]]
        elif quantities == 'layer-line correlation':
            labels += SpringDataBase().get_labels_from_table(SegmentTable, 'ccc_layer')
            labels += ['Selected ' + SpringDataBase().get_labels_from_table(SegmentTable, 'ccc_layer')[0]]
        elif quantities == 'defocus':
            labels += SpringDataBase().get_labels_from_table(SegmentTable, 'avg_defocus')
            labels += ['Selected ' + SpringDataBase().get_labels_from_table(SegmentTable, 'avg_defocus')[0]]
        elif quantities == 'astigmatism':
            labels += SpringDataBase().get_labels_from_table(SegmentTable, 'astigmatism')
            labels += ['Selected ' + SpringDataBase().get_labels_from_table(SegmentTable, 'astigmatism')[0]]
        elif quantities == 'classes':
            labels += SpringDataBase().get_labels_from_table(SegmentTable, 'class_id')
            labels += ['Selected ' + SpringDataBase().get_labels_from_table(SegmentTable, 'class_id')[0]]
        elif quantities == 'class_models':
            labels += SpringDataBase().get_labels_from_table(SegmentTable, 'class_model_id')
            labels += ['Selected ' + SpringDataBase().get_labels_from_table(SegmentTable, 'class_model_id')[0]]
        
        return labels
    
    
    def get_segment_ids_according_to_set_size(self, session, set_size, each_plot_set):
        if set_size == 'helix':
            each_helix_segments = session.query(SegmentTable).filter(SegmentTable.helix_id == each_plot_set.id).\
            order_by(SegmentTable.id).all()
        elif set_size == 'micrograph':
            each_helix_segments = session.query(SegmentTable).filter(SegmentTable.mic_id == each_plot_set.id).\
            order_by(SegmentTable.id).all()
        elif set_size == 'data_set':
            each_helix_segments = session.query(SegmentTable).order_by(SegmentTable.id).all()
            
        each_helix_stack_ids = [each_helix_segment.stack_id for each_helix_segment in each_helix_segments]
        
        return each_helix_stack_ids, each_helix_segments
    

    def generate_set_id_label_for_plot_and_filename(self, set_size, each_plot_set):
        if set_size == 'helix':
            each_set_id_label = 'helixid{0:05}_{1}'.format(each_plot_set.id, each_plot_set.helix_name)
        elif set_size == 'micrograph':
            each_set_id_label = 'micid{0:04}_{1}'.format(each_plot_set.id, each_plot_set.micrograph_name)
            mic_path = os.path.join(each_plot_set.dirname, each_plot_set.micrograph_name)
            if os.path.exists(mic_path):
                os.symlink(mic_path, each_set_id_label)
        elif set_size == 'data_set':
            each_set_id_label = 'dataset_all'
            
        return each_set_id_label
    

    def get_specified_quantities(self, each_plot_set_string, set_size, quantities, xy_quantities, each_helix_stack_ids,
    each_helix_segments, overlapping_segments):

        all_segments_included = np.array([True if each_helix_segment.stack_id in overlapping_segments \
                                         else False for each_helix_segment in each_helix_segments], dtype=bool)

        if quantities == 'coordinates':
            x_coord = np.array([each_helix_segment.x_coordinate_A for each_helix_segment in each_helix_segments])
            y_coord = np.array([each_helix_segment.y_coordinate_A for each_helix_segment in each_helix_segments])
            picked_x_coord = [each_helix_segment.picked_x_coordinate_A for each_helix_segment in each_helix_segments]
            picked_y_coord = [each_helix_segment.picked_y_coordinate_A for each_helix_segment in each_helix_segments]
            self.check_segmenting_values_and_raise_error(quantities, y_coord)
            
            x_coord[all_segments_included == False] = np.nan 
            y_coord[all_segments_included == False] = np.nan 
            xy_quantities.append((each_plot_set_string, np.vstack([each_helix_stack_ids, x_coord, y_coord,
            picked_x_coord, picked_y_coord])))
        else:
            distances = [each_helix_segment.distance_from_start_A for each_helix_segment in each_helix_segments]
            if quantities == 'in-plane_rotation':
                y_plot = [each_helix_segment.inplane_angle for each_helix_segment in each_helix_segments]
                self.check_segmenting_values_and_raise_error(quantities, y_plot)
            elif quantities == 'curvature':
                y_plot = [each_helix_segment.curvature for each_helix_segment in each_helix_segments]
                self.check_segmenting_values_and_raise_error(quantities, y_plot)
            elif quantities == 'defocus':
                y_plot = [each_helix_segment.avg_defocus for each_helix_segment in each_helix_segments]
                self.check_ctf_values_and_raise_error(quantities, y_plot)
            elif quantities == 'astigmatism':
                y_plot = [each_helix_segment.astigmatism for each_helix_segment in each_helix_segments]
                self.check_ctf_values_and_raise_error(quantities, y_plot)
                self.check_ctf_values_and_raise_error(quantities, y_plot)
            elif quantities == 'layer-line correlation':
                y_plot = [each_helix_segment.ccc_layer for each_helix_segment in each_helix_segments]
                self.check_layerline_values_and_raise_error(quantities, y_plot)
            elif quantities == 'classes':
                y_plot = [each_helix_segment.class_id for each_helix_segment in each_helix_segments]
                self.check_class_values_and_raise_error(quantities, y_plot)
            elif quantities == 'class_models':
                y_plot = [each_helix_segment.class_model_id for each_helix_segment in each_helix_segments]
                self.check_class_model_values_and_raise_error(quantities, y_plot)
                
            sel_y_plot = np.array(y_plot, dtype=float)
            if set_size == 'helix':
                sel_y_plot[all_segments_included == False] = np.nan

                xy_quantities.append((each_plot_set_string, np.vstack([each_helix_stack_ids, distances, y_plot,
                sel_y_plot])))
            elif set_size == 'data_set':
                each_helix_stack_ids = np.array(each_helix_stack_ids)
                sel_y_plot = sel_y_plot[all_segments_included == True]
                each_helix_stack_ids = each_helix_stack_ids[all_segments_included == True]
                xy_quantities.append((each_plot_set_string, np.vstack([each_helix_stack_ids, sel_y_plot])))
                
        return xy_quantities
    

    def prepare_log_string(self, labels, xy_quantities):
        msg = ''
        for each_helix, each_quant in xy_quantities:
            msg += ('\n' + tabulate(each_quant.T, labels))
        
        return msg


    def get_quantities_per_set(self, spring_path, quantities):
        self.log.fcttolog()
        shutil.copy(spring_path, 'spring.db')
        
        session, plot_sets = self.get_plot_set(self.set_size, self)

        self.curvature_range, self.ccc_layer_range = SegmentSelect().convert_curvature_ccc_layer_range('spring.db',
        self.straightness_selection, self.curvature_range_perc, self.ccc_layer_selection, self.ccc_layer_range_perc)
        
        combined_included_segments, excluded_segment_counts  = \
        SegmentSelect().filter_non_orientation_parameters_based_on_selection_criteria(self)
        
        labels = self.get_labels_according_to_quantities(self.set_size, quantities)
        xy_quantities = []
        for each_plot_set in plot_sets:
            each_helix_stack_ids, each_helix_segments = self.get_segment_ids_according_to_set_size(session,
            self.set_size, each_plot_set)
            
            if len(set(each_helix_stack_ids).intersection(combined_included_segments)) != 0:
                each_set_id_label = self.generate_set_id_label_for_plot_and_filename(self.set_size, each_plot_set)
                overlapping_segments = list(set(each_helix_stack_ids).intersection(combined_included_segments))
                
                xy_quantities = self.get_specified_quantities(each_set_id_label, self.set_size, quantities,
                xy_quantities, each_helix_stack_ids, each_helix_segments, overlapping_segments)
        
        msg = self.prepare_log_string(labels, xy_quantities)
        
        self.log.ilog('The following quantities have been extracted and will be plotted:\n{0}'.format(msg))
                    
        return labels, xy_quantities
    
    
class SegmentPlotVisualize(SegmentPlotPrepare):
    def determine_position_of_point_within_frame(self, frame_bottom_left, frame_top_right, x_quantity, y_quantity,
    x_lim, y_lim):
        """
        >>> from spring.segment2d.segmentplot import SegmentPlot
        >>> s = SegmentPlot()
        >>> x_quant = np.arange(5.0)
        >>> y_quant = np.arange(5.0, 10.0)
        >>> s.determine_position_of_point_within_frame([0.1, 0.2], [0.9, 0.8], x_quant, y_quant, (0.0, 4.0), (5.0, 9.0)) 
        (array([ 0.1,  0.3,  0.5,  0.7,  0.9]), array([ 0.2       ,  0.26666667,  0.33333333,  0.4       ,  0.46666667]))
        """
        x_quantity = np.array(x_quantity)
        y_quantity = np.array(y_quantity)
        
        scaled_x_quant = (x_quantity - x_lim[0]) / x_lim[-1]
        scaled_y_quant = (y_quantity - y_lim[0]) / y_lim[-1]
        
        adjusted_scaled_x_quant = frame_bottom_left[0] + scaled_x_quant * (frame_top_right[0] - frame_bottom_left[0]) 
        adjusted_scaled_y_quant = frame_bottom_left[1] + scaled_y_quant * (frame_top_right[1] - frame_bottom_left[1]) 
        
        return adjusted_scaled_x_quant, adjusted_scaled_y_quant
    
        
    def add_stack_ids_to_upper_x_axis(self, ax1, seg_ids, x_quantity, y_quantity, labels):
        ax11 = ax1.twiny()
        ax11.plot(x_quantity/x_quantity[1], y_quantity, '.', markersize=0.01)
        ax11.set_xlabel('Stack_id along helix', fontsize=8)
        ax1.set_xlim(min(x_quantity), max(x_quantity))
        ax11.set_xlim(min(seg_ids), max(seg_ids))
        if labels[-1].endswith('in-plane rotation angle (degrees)'.title()):
            ax1.set_ylim(0, 360)
            ax11.set_ylim(0, 360)
        elif labels[-1].endswith('out-of-plane rotation angle (degrees)'.title()):
            ax1.set_ylim(-max(abs(y_quantity)), max(abs(y_quantity)))
            ax11.set_ylim(-max(abs(y_quantity)), max(abs(y_quantity)))
        elif min(y_quantity) < max(y_quantity):
            ax1.set_ylim(min(y_quantity), max(y_quantity))
            ax11.set_ylim(min(y_quantity), max(y_quantity))
        
        return ax1, ax11
    

    def add_legend_at_best_location(self, ax1):
        ax1.legend(loc=0, ncol=1, prop=FontProperties(size='small'))
        
        return ax1


    def set_legend_and_xlabel(self, label, ax1):
        ax1 = self.add_legend_at_best_location(ax1)
        ax1.set_xlabel(label, fontsize=8)
        
        return ax1 
    

    def plot_xy_coordinates(self, labels, set_id_label, xy_data, fig):
        ax1 = fig.add_subplot(111)
        ax1.set_aspect('equal')
        ax1.apply_aspect()
        
        if xy_data.shape[0] == 5:
            seg_ids, x_coord, y_coord, picked_x_coord, picked_y_coord = np.split(xy_data.ravel(), xy_data.shape[0])
        elif xy_data.shape[0] == 7:
            seg_ids, x_coord, y_coord, picked_x_coord, picked_y_coord, sel_x_coord, sel_y_coord = \
            np.split(xy_data.ravel(), xy_data.shape[0])
        
        ax1.set_title('{0}{1}'.format(set_id_label[0].capitalize(), set_id_label[1:]))
        ax1.plot(x_coord, y_coord, '.', label=labels[1])
        ax1.plot(picked_x_coord.ravel(), picked_y_coord.ravel(), 'x', label=labels[3])
        if xy_data.shape[0] == 7: 
            ax1.plot(sel_x_coord, sel_y_coord, 'o', markerfacecolor='none', label=labels[5], fillstyle='bottom')
        ax1 = self.set_legend_and_xlabel(labels[1], ax1)
        ax1.set_xticks([min(min(picked_x_coord), min(x_coord)), max(max(picked_x_coord), max(x_coord))])
        ax1.set_yticks([min(min(picked_y_coord), min(y_coord)), max(max(picked_y_coord), max(y_coord))])
        [each_tick.set_fontsize(8) for each_tick in ax1.get_xticklabels() + ax1.get_yticklabels()]
        ax1.set_ylabel(labels[2], fontsize=8)

        if os.path.exists(set_id_label):
            mic = EMData()
            mic.read_image(set_id_label)
            xdim, ydim = mic.get_xsize(), mic.get_ysize()
            mic = image_decimate(mic, int(round(6.0 / self.pixelsize)), fit_to_fft=False)
            mic_np = np.copy(EMNumPy.em2numpy(mic))

            ax1.imshow(mic_np, cmap='gray', interpolation='nearest', origin='lower', 
                       extent=[0, xdim * self.pixelsize, 0, ydim * self.pixelsize])
            
        return fig
    

    def add_avg_and_stdev_for_label(self, y_quantity):
        y_quantity = np.array(y_quantity)
        y_quantity = y_quantity[(y_quantity > 0) ^ (y_quantity <= 0)]
        avg_stdev = 'Mean: {0:.3}\nstdev: {1:.3}'.format(np.average(y_quantity), np.std(y_quantity))
        
        return avg_stdev
    

    def plot_y_histogram(self, labels, xy_data, set_size, fig):
        seg_ids, y_quantity = np.split(xy_data.ravel(), xy_data.shape[0])
        
        ax1 = fig.add_subplot(111)
        avg_std = self.add_avg_and_stdev_for_label(y_quantity)
        if min(y_quantity) < max(y_quantity):
            ax1.hist(y_quantity, bins=40, label=avg_std)
        [each_tick.set_fontsize(8) for each_tick in ax1.get_xticklabels() + ax1.get_yticklabels()]
        ax1 = self.set_legend_and_xlabel(labels[1], ax1)
        ax1.set_ylabel('Number of segments per {0}'.format(set_size))
        
        return fig


    def make_distance_plot_including_histogram(self, diag_plot, labels, seg_ids, xy_data, set_size, fig, x_quantity,
    y_quantity, sel_y_quantity):
        ax1 = diag_plot.plt.subplot2grid((2, 4), (0, 0), colspan=3, rowspan=2)
        ax1.plot(x_quantity, y_quantity, 'x', label=labels[2])
        if xy_data.shape[0] >= 4:
            ax1.plot(x_quantity, sel_y_quantity, 'o', markerfacecolor='none', label=labels[3])
        ax1.set_xlabel(labels[1], fontsize=8)
        ax1.set_ylabel(labels[2], fontsize=8)
        ax1 = self.set_legend_and_xlabel(labels[1], ax1)
        ax1, ax11 = self.add_stack_ids_to_upper_x_axis(ax1, seg_ids, x_quantity, y_quantity, labels)
        ax2 = diag_plot.plt.subplot2grid((2, 4), (0, 3), colspan=1, rowspan=2)
        if xy_data.shape[0] >= 4:
            avg_std = self.add_avg_and_stdev_for_label(sel_y_quantity)
        else:
            avg_std = self.add_avg_and_stdev_for_label(y_quantity)

        if min(y_quantity) < max(y_quantity):
            ax2.hist(y_quantity, bins=40, orientation='horizontal', label=avg_std)
        ax2.set_xlabel('Number of segments \nper {0}'.format(set_size), fontsize=8)
        ax2 = self.add_legend_at_best_location(ax2)
        ax2.set_ylim(ax1.get_ylim())
        ax2.set_yticks([])
        
        [each_tick.set_fontsize(8) for each_tick in ax1.get_xticklabels() + ax1.get_yticklabels() + \
        ax11.get_xticklabels() + ax2.get_xticklabels()]
        
        return ax1, ax2


    def plot_distance_along_y_including_local_average(self, diag_plot, labels, seg_ids, xy_data, set_size, fig):
        seg_ids, x_quantity, y_quantity, sel_y_quantity, avg_y_quantity = np.split(xy_data.ravel(), xy_data.shape[0])
        
        ax1, ax2 = self.make_distance_plot_including_histogram(diag_plot, labels, seg_ids, xy_data, set_size, fig,
        x_quantity, y_quantity, sel_y_quantity)
        
        ax1.plot(x_quantity, avg_y_quantity, ':', label=labels[-1])
        ax1 = self.add_legend_at_best_location(ax1)
        
        return fig
    
    
    def plot_distance_along_y(self, diag_plot, labels, seg_ids, xy_data, set_size, fig):
        if xy_data.shape[0] == 3:
            seg_ids, x_quantity, y_quantity = np.split(xy_data.ravel(), xy_data.shape[0])
            sel_y_quantity = None
        elif xy_data.shape[0] == 4:
            seg_ids, x_quantity, y_quantity, sel_y_quantity = np.split(xy_data.ravel(), xy_data.shape[0])
        
        self.make_distance_plot_including_histogram(diag_plot, labels, seg_ids, xy_data, set_size, fig, x_quantity,
        y_quantity, sel_y_quantity)
        
        return fig
    

    def choose_subplot_according_to_dimension(self, diag_plot, labels, set_id_label, segid_xy, set_size, fig):
        if segid_xy.shape[0] == 5 and not labels[-1].startswith('Local Average') or segid_xy.shape[0] == 7:
            fig = self.plot_xy_coordinates(labels, set_id_label, segid_xy, fig)
        elif segid_xy.shape[0] == 5 and labels[-1].startswith('Local Average'):
            fig = self.plot_distance_along_y_including_local_average(diag_plot, labels, set_id_label, segid_xy,
            set_size, fig)
        elif segid_xy.shape[0] == 3 or segid_xy.shape[0] == 4:
            fig = self.plot_distance_along_y(diag_plot, labels, set_id_label, segid_xy, set_size, fig)
        elif segid_xy.shape[0] == 2:
            fig = self.plot_y_histogram(labels, segid_xy, set_size, fig)
            
        return fig
            

    def print_progress_statement_if_interactive(self, xy_quantities, batch_mode):
        if not batch_mode:
            progress_statement = '{0} plots are being prepared before they can be browsed interactively.'.\
            format(len(xy_quantities))
            print progress_statement
            self.log.ilog(progress_statement)


    def plot_data_on_figure(self, xy_quantities, labels):
        self.log.fcttolog()
        self.print_progress_statement_if_interactive(xy_quantities, self.batch_mode)
        
        figures = []
        segmentplot_plot = DiagnosticPlot()
        self.log.plog(20)
        for quant_index, (each_set_id_label, each_segid_xy) in enumerate(xy_quantities):
            print each_set_id_label
            fig = segmentplot_plot.create_next_figure()
            
            filename = '{0}_{1}{2}'.format(os.path.splitext(self.outfile_prefix)[0],
            os.path.splitext(each_set_id_label)[0], os.path.splitext(self.outfile_prefix)[-1])
            
            if self.batch_mode:
                fig = segmentplot_plot.add_header_and_footer(self.feature_set, self.infile, filename)
        
            fig = self.choose_subplot_according_to_dimension(segmentplot_plot, labels, each_set_id_label,
            each_segid_xy, self.set_size, fig)
                    
            if self.batch_mode:
                fig.savefig(filename)
            elif not self.batch_mode:
                figures.append(fig)
            self.log.plog(70 * (quant_index + 1) / len(xy_quantities) + 20)
            self.log.ilog('{0}{1} plot was generated.'.format(each_set_id_label[0].capitalize(), each_set_id_label[1:]))
        
        return figures
            
            
class SegmentPlotGui(QWidget):

    def __init__(self, feature_set, figures, tables, labels, hints = None, parent = None):
        QWidget.__init__(self, parent)
        
        self = SpringCommon().setup_spring_page_top(self, feature_set)
        self.figures = figures
        self.tables = tables
        self.labels = labels
        
        self.graph_items = ['{0}{1}'.format(each_table[0][0].capitalize(), each_table[0][1:]) for each_table in tables]
        
        self.stackedComboBox = QComboBox()
        self.stackedComboBox.addItems(self.graph_items)
        if hints is not None:
            self.stackedComboBox.setToolTip(hints)
        self.layout.addWidget(self.stackedComboBox, 0, 3, 1, 1)
        
        self.stackedWidget = QStackedWidget()
        for each_plot_id, each_figure in enumerate(self.figures):
            self.splitter = QSplitter(Qt.Horizontal)
            self.data_table = SpringDataExploreTablePane(self.tables[each_plot_id][1], self.labels[each_plot_id])
            self.splitter.addWidget(self.data_table)
            self.figure = SpringDataExplore(self.figures[each_plot_id])
            self.splitter.addWidget(self.figure)
            
            self.stackedWidget.addWidget(self.splitter)
        
        self.connect(self.stackedComboBox, SIGNAL('currentIndexChanged(int)'), self.stackedWidget.setCurrentIndex)
        self.layout.addWidget(self.stackedWidget, 2, 0, 2, 5)
        
        self.setLayout(self.layout)
        
            
class SegmentPlot(SegmentPlotVisualize):
        
    def launch_interactive_plot_gui(self, feature_set, figures, tables, labels, hints=None):
        app = QApplication(sys.argv)
        symexplor = SegmentPlotGui(feature_set, figures, tables, labels, hints)
        symexplor.show()
        app.exec_()

        
    def plot_desired_quantities(self):
        labels, xy_quantities = self.get_quantities_per_set(self.infile, self.quantities)
        figures = self.plot_data_on_figure(xy_quantities, labels)
        if not self.batch_mode:
            all_labels = [labels for each_table in xy_quantities]
            self.launch_interactive_plot_gui(self.feature_set, figures, xy_quantities, all_labels)
        
        self.log.endlog(self.feature_set)
        
        
def main():
    # Option handling
    parset = SegmentPlotPar()
    mergeparset = OptHandler(parset)
    ######## Program
    stack = SegmentPlot(mergeparset)
    stack.plot_desired_quantities()

if __name__ == '__main__':
    main()



#!/usr/bin/env python
"""
Test module to check segrefine3dgrid
"""
from glob import glob
from multiprocessing import cpu_count
import os
import shutil
from spring.csinfrastr.csdatabase import SpringDataBase, grid_base, GridRefineTable
from spring.csinfrastr.csproductivity import TestingDirectory
from spring.segment3d.segrefine3dgrid import SegRefine3dGridPar, SegRefine3dGrid
from spring.tests.csinfrastr.test_csreadinput import EndToEndTest
from spring.tests.segment3d.test_segmentrefine3d import TestSegmentRefine3dSetup, TestSegmentRefine3dMain


class TestSegRefine3dGridSetup(TestSegmentRefine3dSetup):
    def prepare_grid_database_to_be_continued(self):
        
        grid_session = SpringDataBase().setup_sqlite_db(grid_base, self.grid_db)
        this_grid, first, second = SegRefine3dGrid().enter_variables_into_grid_database(self.prim_var, self.sec_var, 
        self.prim_range, self.sec_range, self.prim_inc, self.sec_inc, 'rise/rotation')
        
        grid_session.add(this_grid)
        
        grid_cycle = GridRefineTable()
        
        grid_cycle.dirname = 'helical_rise_or_pitch_8_helical_rotation_or_number_of_units_per_turn_40'
        grid_cycle.primary_value = 8.0
        grid_cycle.secondary_value = 40.0
        
        grid_cycle.fsc_0143 = 12.0
        grid_cycle.fsc_05 = 18.0
        grid_cycle.amp_correlation = 0.7
        grid_cycle.variance = 0.3
        grid_cycle.xshift_error = 12.5
        grid_cycle.inplane_error = 2.0
        grid_cycle.outofplane_error = 5.0
        grid_cycle.excluded_inplane_ratio = 0.1
        grid_cycle.em_files_2d = ['this_file.hdf']
        grid_cycle.em_files_3d = ['that_file.hdf']
        grid_session.add(grid_cycle)
        
        grid_session.commit()
        
        
    def setup_of_common_sr3dg_non_mpi_parameters(self):
        self.feature_set = SegRefine3dGridPar()
        self.input_stack = 'test_stack.hdf'
        self.prim_var = 'helical_rise_or_pitch'
        self.sec_var = 'helical_rotation_or_number_of_units_per_turn'
        self.grid_db = os.path.join(os.pardir, 'grid.db')
        self.prim_range = ((8.0, 12.0))
        self.sec_range = ((40.0, 60.0))
        self.prim_inc = 2.0
        self.sec_inc = 10.0
        
        self.feature_set.parameters['First parameter']=self.prim_var
        self.feature_set.parameters['Second parameter']=self.sec_var
        self.feature_set.parameters['Lower and upper limit first parameter']=self.prim_range
        self.feature_set.parameters['Lower and upper limit second parameter']=self.sec_range
        self.feature_set.parameters['First and second parameter increment']=((self.prim_inc, self.sec_inc))
        self.feature_set.parameters['Subgrid option']=False
        self.feature_set.parameters['Part and number of subgrids']=(3, 3)
        self.feature_set.parameters['Grid continue option']=False
        self.feature_set.parameters['Grid database']=self.grid_db
        self.setup_main_segmentrefine3d_parameters()
        
class TestSegRefine3dGrid(TestSegRefine3dGridSetup, SegRefine3dGrid):
    def setup_srg3d(self):
        self.setup_of_common_sr3dg_non_mpi_parameters()
        self.feature_set.parameters['MPI option']=False
        self.feature_set.parameters['Number of CPUs']=min(3, cpu_count())
        self.feature_set.parameters['Temporary directory']=os.path.abspath(os.curdir)
        
        self.prepare_apolar_helix = False
        self.generate_input_stack_from_helix()
        
        super(TestSegRefine3dGridSetup, self).__init__(self.feature_set)
        
    def setup(self):
        self.testingdir = TestingDirectory(os.path.splitext(os.path.basename(__file__))[0])
        self.testingdir.create()
        
        self.setup_srg3d()

    def teardown(self):
        parfiles = glob('*par')
        logfiles = glob('*log')
        directory = 'helical_rise_or_pitch'
        dirfiles = glob('{dir}*{sep}*'.format(dir=directory, sep=os.sep))
        for each_file in logfiles + parfiles + dirfiles:
            os.remove(each_file)
        for each_dir in glob('{0}*'.format(directory)):
            shutil.rmtree(each_dir)
           
        os.remove(self.input_stack)
        os.remove(os.path.join(os.path.pardir, 'spring.db'))
        for each_file in glob('gri*.db'):
            os.remove(each_file)
         
        if self.feature_set.parameters['Continue refinement option'] is True:
            self.tsr3d.testingdir.remove()
        self.testingdir.remove()
            
        
class TestSegRefine3dGridMain(TestSegRefine3dGrid):
    def do_test_case_srg1(self):
        """
        Test standard Segrefine3dgrid 
        """
        self.launch_segmentrefine3d_jobs()


class TestSegRefine3dGridMore(TestSegRefine3dGrid):
    def to_test_case_srg2(self):
        """
        Test Segrefine3dgrid using subgrid option
        """
        self.feature_set.parameters['Subgrid option']=True
          
        super(TestSegRefine3dGridSetup, self).__init__(self.feature_set)
        self.launch_segmentrefine3d_jobs()
         
         
    def to_test_case_srg3(self):
        """
        Test Segrefine3dgrid using continue grid option
        """
        self.feature_set.parameters['Subgrid option']=False
        self.feature_set.parameters['Grid continue option']=True
          
        self.prepare_grid_database_to_be_continued()
          
        super(TestSegRefine3dGridSetup, self).__init__(self.feature_set)
          
        self.launch_segmentrefine3d_jobs()
        os.remove(self.grid_db)
          
          
    def to_test_case_srg4(self):
        """
        Test Segrefine3dgrid using continue refinement option
        """
        self.feature_set.parameters['Continue refinement option']=True
        self.feature_set.parameters['refinement.db file']='refinement006.db'
        self.feature_set.parameters['Number of iterations']=3
        os.remove(os.path.join(os.pardir, 'spring.db'))
         
        super(TestSegRefine3dGridSetup, self).__init__(self.feature_set)
        self.tsr3d = TestSegmentRefine3dMain()
        self.tsr3d.setup()
        self.tsr3d.do_test_case_sr1()
         
        self.launch_segmentrefine3d_jobs()
        self.tsr3d.clean_files_from_iterations()
        
          
    def to_test_case_srg5(self):
        """
        Test Segrefine3dgrid using swapped primary and secondary variables
        """
        self.feature_set.parameters['First parameter']=self.sec_var
        self.feature_set.parameters['Second parameter']=self.prim_var
        self.feature_set.parameters['Lower and upper limit first parameter']=self.sec_range
        self.feature_set.parameters['Lower and upper limit second parameter']=self.prim_range
        self.feature_set.parameters['First and second parameter increment']=((self.sec_inc, self.prim_inc))
 
        super(TestSegRefine3dGridSetup, self).__init__(self.feature_set)
        self.launch_segmentrefine3d_jobs()
         

class TestSegRefine3dGridEndToEnd(TestSegRefine3dGrid):
    def do_end_to_end_test_sr3d_inputfile(self):
        EndToEndTest().do_end_to_end_inputfile(self.feature_set)
 

    def do_end_to_end_test_sr3d_prompt(self):
        EndToEndTest().do_end_to_end_prompt(self.feature_set)
        
        
def main():
    tsr3d = TestSegRefine3dGridMain()
    tsr3d.setup()
    tsr3d.do_test_case_srg1()
        
if __name__ == '__main__':
    main()



#!/usr/bin/env python
"""
Test module to check segmentctfapply
"""
from multiprocessing import cpu_count
from spring.csinfrastr.csproductivity import TestingDirectory
from spring.segment2d.segmentctfapply import SegmentCtfApply, SegmentCtfApplyPar
from spring.tests.micprgs.test_micctfdetermine import TestMicCtfDetermineMain
from spring.tests.segment2d.test_segment import TestSegmentMain
import os
import shutil
# from spring.tests.csinfrastr.test_csreadinput import EndToEndTest

        
class TestSegmentCtfApplyPreparation(object):
    def generate_helix_stack_from_micrograph(self):
        self.tscd = TestMicCtfDetermineMain()
        self.tscd.setup()
        self.tscd.do_test_case_scd1()
        
        shutil.copy('spring.db', os.path.join(os.pardir, 'spring.db'))
        self.tscd.teardown()
        
        self.tseg = TestSegmentMain()
        self.tseg.micrograph_test_file = self.tscd.micrograph_file
        self.tseg.setup()
        self.tseg.do_test_case_seg1()
        
        os.remove(self.tseg.micrograph_test_file)
        shutil.move('spring.db', os.path.join(os.pardir, 'segment_spring.db'))
        shutil.copy(self.image_inp_stack, os.path.join(os.pardir, 'segmented_stack.hdf'))
        self.tseg.teardown()
        shutil.move('segmented_stack.hdf', self.image_inp_stack)
        

class TestSegmentCtfApply(SegmentCtfApply, TestSegmentCtfApplyPreparation):
        
    def setup_of_common_non_mpi_parameters(self):
        self.feature_set = SegmentCtfApplyPar()
        
        self.image_inp_stack = 'test_output_stack.hdf'
        self.image_out_stack = 'test_segments_ctf.hdf'
        self.test_ctfvolume = 'test_ctfvolume.hdf'
        
        self.feature_set.parameters['Image input stack'] = self.image_inp_stack
        self.feature_set.parameters['Image output stack'] = self.image_out_stack
        
        self.feature_set.parameters['spring.db file']='segment_spring.db'#os.path.join(os.pardir, 'spring.db')
            
        self.feature_set.parameters['Pixel size in Angstrom']=5.0
        self.feature_set.parameters['CTFFIND or CTFTILT'] = 'ctftilt'
        self.feature_set.parameters['convolve or phase-flip'] = 'convolve'
        self.feature_set.parameters['Astigmatism correction'] = True
        
    def setup(self):
        self.testingdir = TestingDirectory(os.path.splitext(os.path.basename(__file__))[0])
        self.testingdir.create()

        self.setup_of_common_non_mpi_parameters()
        self.feature_set.parameters['MPI option']=False
        self.feature_set.parameters['Number of CPUs']=min(8, cpu_count())
        self.feature_set.parameters['Temporary directory']=os.path.abspath(os.curdir)
        
        self.generate_helix_stack_from_micrograph()
        super(TestSegmentCtfApply, self).__init__(self.feature_set)
        
    def teardown(self):
        os.remove(self.image_inp_stack)
        os.remove(self.image_out_stack)
        
        self.testingdir.remove()
    

class TestSegmentCtfApplyMain(TestSegmentCtfApply):
    def do_test_case_sca1(self):
        self.apply_ctf_to_segment_stack()


class TestSegmentCtfApplyMore(TestSegmentCtfApply):
    def do_test_case_sca2(self):
        self.feature_set.parameters['CTFFIND or CTFTILT'] = 'ctffind'
        super(TestSegmentCtfApply, self).__init__(self.feature_set)
        
        self.apply_ctf_to_segment_stack()
        
    def do_test_case_sca3(self):
        self.feature_set.parameters['convolve or phase-flip'] = 'phase-flip'
        super(TestSegmentCtfApply, self).__init__(self.feature_set)
        
        self.apply_ctf_to_segment_stack()
        
# class TestSegmentCtfApplyEndToEnd(TestSegmentCtfApply):
#     def do_end_to_end_test_sca3_inputfile(self):
#         EndToEndTest().do_end_to_end_inputfile(self.feature_set)
#  
#     def do_end_to_end_test_sca4_prompt(self):
#         EndToEndTest().do_end_to_end_prompt(self.feature_set)
#         
#         
# class TestSegmentCtfApplyMpi(TestSegmentCtfApply):
#     def do_end_to_end_test_sa_inputfile_with_MPI(self):
#         self.feature_set.parameters['MPI option'] = True
#         self.feature_set.parameters['Number of CPUs']=min(8, cpu_count())
#         
#         EndToEndTest().do_end_to_end_inputfile(self.feature_set)


def main():
    tsca = TestSegmentCtfApplyMain()
    tsca.setup()
    tsca.do_test_case_sca1()
        
if __name__ == '__main__':
    main()
